# Copyright Metatype OÜ under the Elastic License 2.0 (ELv2). See LICENSE.md for usage.

import ast
import inspect
import os
from typing import Any, Dict, List, Optional, Tuple

from astunparse import unparse
from attrs import field, frozen
from frozendict import frozendict

from typegraph import effects
from typegraph.effects import Effect
from typegraph.graph.builder import Collector
from typegraph.graph.nodes import Node
from typegraph.runtimes.base import Materializer, Runtime
from typegraph.utils.attrs import SKIP, always


@frozen
class DenoRuntime(Runtime):
    worker: str = field(kw_only=True, default="default")
    allow_net: Tuple[str, ...] = field(
        kw_only=True, factory=tuple, metadata={SKIP: True}
    )
    permissions: Dict[str, Any] = field(factory=frozendict, init=False)
    runtime_name: str = always("deno")

    def __attrs_post_init__(self):
        permissions = {}
        if len(self.allow_net) > 0:
            if "*" in self.allow_net:
                permissions["net"] = True
            else:
                permissions["net"] = self.allow_net
        object.__setattr__(self, "permissions", frozendict(permissions))


class LambdaCollector(ast.NodeTransformer):
    @classmethod
    def collect(cls, function):
        source = inspect.getsource(function).lstrip()
        tree = ast.parse(source)
        ret = cls()
        ret.visit(tree)
        return ret.lambdas

    def __init__(self):
        super().__init__()
        self.lambdas = []

    def visit_Lambda(self, node):
        self.lambdas.append(unparse(node).strip())


# Inlined fuction
@frozen
class FunMat(Materializer):
    fn_expr: Optional[str] = field(default=None, metadata={SKIP: True})

    # a script that assigns a function expression into the variable _my_lambda
    script: Optional[str] = field(kw_only=True, default=None)
    runtime: DenoRuntime = field(kw_only=True, factory=DenoRuntime)
    materializer_name: str = field(default="function", init=False)
    effect: Effect = field(kw_only=True, default=effects.none())

    @classmethod
    def from_lambda(cls, function, runtime=DenoRuntime()):
        lambdas = LambdaCollector.collect(function)
        assert len(lambdas) == 1
        raise NotImplementedError

    def __attrs_post_init__(self):
        if self.fn_expr is None:
            if self.script is None:
                raise Exception("you must give the script or a function expression")
        else:
            if self.script is not None:
                raise Exception(
                    "you must only give either the script or a function expression"
                )
            object.__setattr__(self, "script", f"var _my_lambda = {self.fn_expr};")


@frozen
class PureFunMat(FunMat):
    effect: Effect = always(effects.none())


@frozen
class PredefinedFunMat(Materializer):
    name: str
    effect: Effect = field(kw_only=True, default=effects.none())
    runtime: DenoRuntime = field(kw_only=True, factory=DenoRuntime)
    materializer_name: str = always("predefined_function")


# Import function from a module
@frozen
class ImportFunMat(Materializer):
    mod: "ModuleMat" = field()
    name: str = field(default="default")
    secrets: Tuple[str] = field(kw_only=True, factory=tuple)
    effect: Effect = field(kw_only=True, default=effects.none())
    runtime: DenoRuntime = field(
        kw_only=True, factory=DenoRuntime
    )  # should be the same runtime as `mod`'s
    materializer_name: str = always("import_function")
    collector_target = always(Collector.materializers)

    @property
    def edges(self) -> List[Node]:
        return super().edges + [self.mod]

    def data(self, collector: Collector) -> dict:
        data = super().data(collector)
        data["data"]["mod"] = collector.index(self.mod)
        return data


@frozen
class ModuleMat(Materializer):
    file: Optional[str] = field(default=None, metadata={SKIP: True})
    secrets: Tuple[str] = field(kw_only=True, factory=tuple)
    code: Optional[str] = field(kw_only=True, default=None)
    runtime: DenoRuntime = field(kw_only=True, factory=DenoRuntime)  # DenoRuntime
    materializer_name: str = always("module")
    effect: Effect = always(effects.none())

    def __attrs_post_init__(self):
        if self.file is None:
            if self.code is None:
                raise Exception(
                    f"you must give source code for the module: {self.file}"
                )
        else:
            if self.code is not None:
                raise Exception("you must only give either source file or source code")

            from typegraph.graph.typegraph import get_absolute_path

            path = get_absolute_path(self.file)
            if os.environ.get("DONT_READ_EXTERNAL_TS_FILES"):
                object.__setattr__(self, "code", f"file:{path}")
            else:
                with open(path) as f:
                    object.__setattr__(self, "code", f.read())

    def imp(
        self, name: str = "default", *, effect: Effect = effects.none(), **kwargs
    ) -> ImportFunMat:
        return ImportFunMat(
            self,
            name,
            runtime=self.runtime,
            secrets=self.secrets,
            effect=effect,
            **kwargs,
        )


@frozen
class IdentityMat(PredefinedFunMat):
    name: str = always("identity")
    effect: Effect = always(effects.none())



# Copyright Metatype OÜ under the Elastic License 2.0 (ELv2). See LICENSE.md for usage.

from typing import Dict, Union

from ordered_set import OrderedSet

from typegraph.graph.nodes import Node, NodeProxy


class Collector:
    types = "types"
    runtimes = "runtimes"
    materializers = "materializers"
    policies = "policies"
    secrets = "secrets"

    collects: Dict[str, OrderedSet]
    frozen = False

    def __init__(self):
        self.collects = {
            f: OrderedSet()
            for f in [
                Collector.types,
                Collector.runtimes,
                Collector.materializers,
                Collector.policies,
                Collector.secrets,
            ]
        }

    def freeze(self):
        self.frozen = True

    def collect(self, node: Node) -> bool:
        # Add `node` to its target collect.
        # Returns:
        #    bool: `True` if the node was added to the collect,
        #          `False` if it had already been added from previous operations

        if self.frozen:
            raise Exception("Frozen collector cannot collect")

        if isinstance(node, NodeProxy):
            node = node.get()

        c = node.collector_target
        if c is None:
            raise Exception("Attempting to collect non collectible node")

        collect = self.collects[c]
        prev_size = len(collect)

        collect.add(node)
        return len(collect) > prev_size

    # returns the index of the node in its collect
    def index(self, node: Union[Node, "NodeProxy"]):
        if not isinstance(node, Node):
            raise Exception(f"expected Node, got {type(node).__name__}")

        if isinstance(node, NodeProxy):
            node = node.get()

        c = node.collector_target
        if c not in self.collects:
            raise Exception(f"Collector target '{c}' not found")

        return self.collects[c].index(node)



# Copyright Metatype OÜ under the Elastic License 2.0 (ELv2). See LICENSE.md for usage.

from attrs import frozen

from typegraph import effects
from typegraph.effects import Effect
from typegraph.runtimes.base import Materializer, Runtime
from typegraph.utils.attrs import always


@frozen
class TypeGraphRuntime(Runtime):
    runtime_name: str = always("typegraph")


@frozen
class TypeMat(Materializer):
    runtime: Runtime = TypeGraphRuntime()
    materializer_name: str = always("getType")
    effect: Effect = always(effects.none())


@frozen
class SchemaMat(Materializer):
    runtime: Runtime = TypeGraphRuntime()
    materializer_name: str = always("getSchema")
    effect: Effect = always(effects.none())


@frozen
class ResolverMat(Materializer):
    runtime: Runtime = TypeGraphRuntime()
    materializer_name: str = always("resolver")
    effect: Effect = always(effects.none())


@frozen
class TypeGateRuntime(Runtime):
    runtime_name: str = always("typegate")


@frozen
class TypeGraphsMat(Materializer):
    runtime: Runtime = TypeGateRuntime()
    materializer_name: str = always("typegraphs")
    effect: Effect = always(effects.none())


@frozen
class TypeGraphMat(Materializer):
    runtime: Runtime = TypeGateRuntime()
    materializer_name: str = always("typegraph")
    effect: Effect = always(effects.none())


@frozen
class AddTypeGraphMat(Materializer):
    runtime: Runtime = TypeGateRuntime()
    materializer_name: str = always("addTypegraph")
    effect: Effect = always(effects.create())


@frozen
class RemoveTypeGraphMat(Materializer):
    runtime: Runtime = TypeGateRuntime()
    materializer_name: str = always("removeTypegraph")
    effect: Effect = always(effects.delete())


# @frozen
# class TypeNodeMat(Materializer):
#     runtime: Runtime = TypeGateRuntime()
#     materializer_name: str = always("typenode")


@frozen
class SerializedTypegraphMat(Materializer):
    runtime: Runtime = TypeGateRuntime()
    materializer_name: str = always("serializedTypegraph")
    effect: Effect = always(effects.none())



# Copyright Metatype OÜ under the Elastic License 2.0 (ELv2). See LICENSE.md for usage.

from attrs import frozen

from typegraph import types as t
from typegraph.runtimes.base import Materializer, Runtime
from typegraph.utils.attrs import always


@frozen
class TemporalRuntime(Runtime):
    """Interacts with Temporal server.

    This runtime is *experimental* state and is subject to change.
    """

    host: str
    runtime_name: str = always("temporal")

    def data(self, collector):
        data = super().data(collector)
        data["data"].update(
            host=self.host,
        )
        return data

    def start_workflow(self, workflow_type: str, arg):
        return t.func(
            t.struct({"workflow_id": t.string(), "args": t.array(arg)}),
            t.string(),
            StartWorkflowMat(self, workflow_type),
        )

    def signal_workflow(self, signal_name: str, arg):
        return t.func(
            t.struct(
                {"workflow_id": t.string(), "run_id": t.string(), "args": t.array(arg)}
            ),
            t.string(),
            SignalWorkflowMat(self, signal_name),
        )

    def query_workflow(self, query_type: str, arg):
        return t.func(
            t.struct(
                {"workflow_id": t.string(), "run_id": t.string(), "args": t.array(arg)}
            ),
            t.string(),
            QueryWorkflowMat(self, query_type),
        )

    def describe_workflow(self):
        return t.func(
            t.struct({"workflow_id": t.string(), "run_id": t.string()}),
            t.string(),
            DescribeWorkflowMat(self),
        )


@frozen
class StartWorkflowMat(Materializer):
    runtime: Runtime
    workflow_type: str
    materializer_name: str = always("start_workflow")


@frozen
class SignalWorkflowMat(Materializer):
    runtime: Runtime
    signal_name: str
    materializer_name: str = always("signal_workflow")


@frozen
class QueryWorkflowMat(Materializer):
    runtime: Runtime
    query_type: str
    materializer_name: str = always("query_workflow")


@frozen
class DescribeWorkflowMat(Materializer):
    runtime: Runtime
    materializer_name: str = always("describe_workflow")



# Copyright Metatype OÜ under the Elastic License 2.0 (ELv2). See LICENSE.md for usage.

from typing import TYPE_CHECKING, Callable, Dict, List, Optional

if TYPE_CHECKING:
    from typegraph import types as t
    from typegraph.graph.builder import Collector
    from typegraph.graph.typegraph import TypeGraph


class Node:
    collector_target: Optional[str]

    def __init__(self, collector_target: Optional[str] = None):
        self.collector_target = collector_target

    @property
    def edges(self) -> List["Node"]:
        return []

    def data(self, collector: "Collector") -> Dict:
        return {}


class NodeProxy(Node):
    graph: "TypeGraph"
    node: str
    after_apply: Optional[Callable[["t.typedef"], "t.typedef"]]

    def __init__(
        self,
        g: "TypeGraph",
        node: str,
        after_apply: Optional[Callable[["t.typedef"], "t.typedef"]] = None,
    ):
        super().__init__()
        self.graph = g
        self.node = node
        self.after_apply = after_apply

    def then(self, then_apply: Callable[["t.typedef"], "t.typedef"]):
        return NodeProxy(
            self.graph, self.node, lambda n: then_apply(self.after_apply(n))
        )

    def get(self) -> "t.typedef":
        tpe = self.graph.type_by_names.get(self.node)
        if tpe is None:
            raise Exception(f"No registered type named '{self.node}'")
        if self.after_apply is None:
            return tpe
        tpe = self.after_apply(tpe)
        self.graph.type_by_names[tpe.name] = tpe
        self.node, self.after_apply = tpe.name, None
        return tpe

    @property
    def edges(self) -> List["Node"]:
        return self.get().edges

    def data(self, collector: "Collector") -> dict:
        return self.get().data(collector)

    @property
    def name(self) -> str:
        return self.node

    def optional(self):
        from typegraph import types as t

        return t.optional(self)



# Copyright Metatype OÜ under the Elastic License 2.0 (ELv2). See LICENSE.md for usage.

# shortcut imports
from typegraph import types as t  # noqa
from typegraph.graph.typegraph import TypeGraph  # noqa

version = "0.0.3-dev.5"



# Copyright Metatype OÜ under the Elastic License 2.0 (ELv2). See LICENSE.md for usage.

from typing import List, Optional, Tuple

from attrs import evolve, frozen

from typegraph import types as t
from typegraph.graph.typegraph import resolve_proxy
from typegraph.providers.prisma.relations import RelationshipRegister

# https://www.prisma.io/docs/reference/api-reference/prisma-schema-reference#model-fields
prisma_types = {
    "string": "String",
    "boolean": "Boolean",
    "integer": "Int",
    # 'BigInt'
    "number": "Float",
    # 'Decimal',
    "datetime": "DateTime",
    # 'Json',
    # 'Bytes'
}


def to_prisma_string(s: str) -> str:
    return f'"{repr(s)[1:-1]}"'


def to_prisma_list(lst: List[str]) -> str:
    return f"[{', '.join(lst)}]"


class PrismaField:
    name: str
    tpe: t.Type
    prisma_name: str
    prisma_type: Optional[str]
    tags: List[str]

    def __init__(self, name: str, tpe: t.Type):
        self.name = name
        self.tpe = tpe
        self.prisma_name = name
        self.prisma_type = None
        self.tags = []


def get_ids(tpe: t.struct) -> Tuple[str, ...]:
    return [
        k for k, v in tpe.props.items() if resolve_proxy(v).runtime_config.get("id")
    ]


@frozen
class SchemaField:
    name: str
    typ: str
    tags: List[str]
    fkeys: List["SchemaField"] = []  # foreign keys
    fkeys_unique: bool = False

    def build(self) -> str:
        return f"{self.name} {self.typ} {' '.join(self.tags)}"


@frozen
class FieldBuilder:
    spec: RelationshipRegister

    def additional_tags(self, typ: t.typedef) -> List[str]:
        tags = []
        if typ.runtime_config.get("id", False):
            tags.append("@id")
        if typ.runtime_config.get("unique", False):
            tags.append("@unique")
        if typ.runtime_config.get("auto", False):
            if typ.type == "integer":
                tags.append("@default(autoincrement())")
            elif typ.type == "string" and typ._format == "uuid":
                tags.append("@default(uuid())")
            else:
                raise Exception(f"'auto' tag not supported for type '{typ.type}'")
        return tags

    def get_type_ids(self, typ: t.struct) -> List[str]:
        return [
            k
            for k, ty in typ.props.items()
            if resolve_proxy(ty).runtime_config.get("id", False)
        ]

    def relation(
        self, field: str, typ: t.struct, rel_name: str
    ) -> [str, List[SchemaField]]:
        references = self.get_type_ids(typ)
        fields = [f"{field}{ref.title()}" for ref in references]

        fkeys = [
            evolve(self.build(ref, typ.props[ref], typ), tags=[], name=f)
            for f, ref in zip(fields, references)
        ]

        name = to_prisma_string(rel_name)
        fields = to_prisma_list(fields)
        references = to_prisma_list(references)

        tag = f"@relation(name: {name}, fields: {fields}, references: {references})"

        return [tag, fkeys]

    def build(self, field: str, typ: t.typedef, parent_type: t.struct) -> SchemaField:
        quant = ""
        if typ.type == "optional":
            quant = "?"
            typ = resolve_proxy(typ.of)
        if typ.type == "array":
            quant = "[]"
            typ = resolve_proxy(typ.of)

        # TODO: default value

        assert typ.type not in ["optional", "array"], "Nested quantifiers not supported"

        tags = []
        fkeys = []
        fkeys_unique = False

        if isinstance(typ, t.string):
            # TODO: enum? json?
            if typ._format == "uuid":
                tags.append("@db.Uuid")  # postgres only
            if typ._format == "date":
                name = "DateTime"
            else:
                name = "String"

        elif isinstance(typ, t.boolean):
            name = "Boolean"

        elif isinstance(typ, t.integer):
            name = "Int"
        elif isinstance(typ, t.number):
            name = "Float"

        else:
            assert typ.type == "object", f"Type f'{typ.type}' not supported"
            name = typ.name

            proxy = self.spec.proxies[parent_type.name].get(field, None)
            assert proxy is not None, f"No proxy for '{parent_type.name}' at '{field}'"
            rel = self.spec.relations[proxy.link_name]

            if isinstance(proxy.get(), t.struct):
                # left side of the relation: the one that has the foreign key defined in
                assert quant == ""
                tag, fkeys = self.relation(field, typ, rel.name)
                tags.append(tag)
                fkeys = fkeys
                fkeys_unique = rel.cardinality.is_one_to_one()
            else:
                # right side of the relation
                tags.append(f"@relation(name: {to_prisma_string(rel.name)})")
            # TODO add additional fields for the foreign keys

        tags.extend(self.additional_tags(typ))

        return SchemaField(
            name=field,
            typ=f"{name}{quant}",
            tags=tags,
            fkeys=fkeys,
            fkeys_unique=fkeys_unique,
        )


def build_model(name: str, spec: RelationshipRegister) -> str:
    fields = []

    # struct
    s = spec.types[name]

    field_builder = FieldBuilder(spec)

    tags = []

    for key, typ in s.props.items():
        typ = resolve_proxy(typ)
        if typ.runtime is not None and typ.runtime != s.runtime:
            continue
        field = field_builder.build(key, typ, s)
        fields.append(field)
        fields.extend(field.fkeys)
        if field.fkeys_unique:
            tags.append(f"@@unique({', '.join((f.name for f in field.fkeys))})")

    # TODO support for multi-field ids and indexes -- to be defined as config on the struct!!

    ids = [field.name for field in fields if "@id" in field.tags]
    assert len(ids) > 0, f"No id field defined in '{name}'"

    if len(ids) > 1:
        # multi-field id
        # ? should require declaration on the struct??
        tags.append(f"@@id([{', '.join(ids)}]")
        for field in fields:
            field.tags.remove("@id")

    # TODO process other struct-level config

    formatted_fields = "".join((f"    {f.build()}\n" for f in fields))
    formatted_tags = "".join(f"    {t}\n" for t in tags)

    return f"""model {name} {{\n{formatted_fields}\n{formatted_tags}}}\n"""



# Copyright Metatype OÜ under the Elastic License 2.0 (ELv2). See LICENSE.md for usage.

import inspect
import re

import black
import httpx
import redbaron
from box import Box


def camel_to_snake(name):
    name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", name).lower()


def upper_first(s):
    return f"{s[0].upper()}{s[1:]}"


def typify(cursor, filter_read_only=False, suffix="", opt=False):
    if (
        not opt
        and "description" in cursor
        and cursor.description.startswith("Optional")
    ):
        return f"t.optional({typify(cursor, filter_read_only, suffix, True)})"

    if "$ref" in cursor:
        return f'g("{cursor["$ref"]}{suffix}")'

    if cursor.type == "string":
        return "t.string()"

    if cursor.type == "boolean":
        return "t.boolean()"

    if cursor.type == "integer":
        # cursor.format = 'int32'
        return "t.integer()"

    if cursor.type == "number":
        # cursor.format = 'double'
        return "t.float()"

    if cursor.type == "any":
        return "t.any()"

    if cursor.type == "array":
        return f't.array({typify(cursor["items"], filter_read_only, suffix)})'

    if cursor.type == "object":
        ret = "t.struct({"

        fields = []
        for f, v in cursor.get("properties", {}).items():
            if filter_read_only or "readOnly" not in v or not v.readOnly:
                fields.append(f'"{f}": {typify(v, filter_read_only, suffix)}')

        if len(fields) == 0:
            # FIXME : accept empty object?
            fields.append('"_": t.optional(t.any())')

        ret += ",".join(fields)
        ret += "}})"
        if "id" in cursor:
            ref = f"{cursor.id}{suffix}"
            ret += f'.named("{ref}")'

        return ret

    raise Exception(f"Unexpect type {cursor}")


def get_effect(method):
    if method == "GET":
        return "effects.none()"
    if method == "POST":
        return "effects.create()"
    if method == "PUT":
        return "effects.upsert()"
    if method == "DELETE":
        return "effects.delete()"
    if method == "PATCH":
        return "effects.update()"
    raise Exception(f"Unsupported HTTP method '{method}'")


def flatten_calls(cursor, hierarchy="", url_prefix=""):
    ret = ""

    if "methods" in cursor:
        for methodName, method in cursor.methods.items():
            inp = ""
            for parameterName, parameter in method.parameters.items():
                if parameterName != "readMask":
                    inp += f'"{parameterName}": {typify(parameter, suffix="In")},'

            out = typify(method.response, suffix="Out")

            effect = get_effect(method.httpMethod)
            mat = f'googleapis.RestMat("{method.httpMethod}", "{url_prefix}{method.path}", effect={effect})'
            ret += f'{hierarchy}{upper_first(methodName)}=t.func(t.struct({{{inp}}}), {out}, {mat}).named("{method.id}"),\n'

    if "resources" in cursor:
        for resourceName, resource in cursor.resources.items():
            ret += flatten_calls(
                resource,
                resourceName
                if hierarchy == ""
                else f"{hierarchy}{upper_first(resourceName)}",
                url_prefix=url_prefix,
            )

    return ret


def codegen(discovery):
    assert discovery.discoveryVersion == "v1"
    assert discovery.protocol == "rest"

    discovery.revision
    discovery.version

    discovery.canonicalName
    discovery.description
    discovery.documentationLink

    cg = ""

    for schema in discovery.schemas.values():
        assert schema.type == "object"

        cg += f'    {camel_to_snake(schema.id)}_in = {typify(schema, filter_read_only=False, suffix="In")}\n'
        cg += f'    {camel_to_snake(schema.id)}_out = {typify(schema, filter_read_only=True, suffix="Out")}\n'

        schema.description

    cg += f"    g.expose({flatten_calls(discovery, url_prefix=discovery.rootUrl)})"

    return cg


def import_googleapis(uri: str, gen: bool) -> None:
    if not gen:
        return

    file = inspect.stack()[1].filename

    with open(file, "r") as f:
        code = redbaron.RedBaron(f.read())

    imports = [
        ["typegraph.providers.google.runtimes", "googleapis"],
        ["typegraph", "t"],
        ["typegraph", "TypeGraph"],
        ["typegraph", "effects"],
    ]

    importer = code.find(
        "atomtrailers", value=lambda x: x.find("name", value="import_googleapis")
    ).find("name", value="True")
    if importer:
        importer.value = "False"

    for frm, imp in imports:
        if not code.find(
            "from_import",
            value=lambda x: x.dumps() == frm,
            targets=lambda x: x.dumps() == imp,
        ):
            code.insert(0, f"from {frm} import {imp}\n")

    discovery = Box(httpx.get(uri).json())

    wth = code.find("with")
    wth.contexts = f'TypeGraph(name="{discovery.name}") as g'
    wth.value = codegen(discovery)

    new_code = black.format_str(code.dumps(), mode=black.FileMode())

    with open(file, "w") as f:
        f.write(new_code)



# Copyright Metatype OÜ under the Elastic License 2.0 (ELv2). See LICENSE.md for usage.

from typing import TYPE_CHECKING, Dict, List, Optional

from attrs import frozen

from typegraph.effects import Effect
from typegraph.graph.builder import Collector
from typegraph.graph.nodes import Node
from typegraph.utils.attrs import always, asdict

if TYPE_CHECKING:
    from typegraph import types as t


@frozen
class Runtime(Node):
    runtime_name: str
    collector_target: Optional[str] = always(Collector.runtimes)

    def data(self, collector: Collector) -> Dict:
        data = asdict(self)
        data.pop("collector_target")
        return {"name": data.pop("runtime_name"), "data": data}

    def get_type_config(self, type: "t.typedef") -> Dict:
        return dict()


@frozen
class Materializer(Node):
    runtime: Runtime
    effect: Effect
    collector_target: Optional[str] = always(Collector.materializers)

    @property
    def edges(self) -> List[Node]:
        return super().edges + [self.runtime]

    def data(self, collector: Collector) -> dict:
        data = asdict(self)
        data.pop("collector_target")
        return {
            "name": data.pop("materializer_name"),
            "runtime": collector.index(data.pop("runtime")),
            "effect": asdict(data.pop("effect")),
            "data": data,
        }



# Copyright Metatype OÜ under the Elastic License 2.0 (ELv2). See LICENSE.md for usage.

import inspect
import itertools
import re
import sys
from ast import literal_eval
from typing import Dict, List, Optional, Set, Tuple

import black
from attrs import define, field, frozen
from redbaron import AtomtrailersNode, NameNode, RedBaron

from typegraph import TypeGraph, t
from typegraph.importers.base.typify import Typify
from typegraph.types import reserved_types

# TODO: detect indentation size/character from file


@define
class Codegen:
    indent: str = field(default=" " * 4)
    indent_level: int = field(default=1)
    res: str = field(default="", init=False)

    def line(self, line: str = "", indent_level=0):
        indent = self.indent * (self.indent_level + indent_level)
        if line == "":
            self.res += "\n"
        else:
            self.res += f"{indent}{line}\n"


class Importer:
    """Base importer class"""

    imports: Set[Tuple[str, str]]
    headers: List[str]  # codegen header lines
    name: str
    types: Dict[str, t.typedef]
    exposed: Dict[str, t.func]
    renames: Dict[str, str]
    tg: TypeGraph

    def __init__(
        self, name: str, *, renames: Dict[str, str] = {}, keep_names: List[str] = []
    ):
        """
        Args:
            `name`: name of the importer
            `renames`: a dictionary mapping original (imported) names to exposed names
            `keep_names`: a list of names to keep as the original (imported)
        """

        self.imports = {
            ("typegraph", "t"),
        }
        self.headers = []

        valid_name = re.sub("\\W", "_", name)
        if valid_name != name:
            raise Exception(
                f"'{name}' is not a valid name for an importer, use '{valid_name}'"
            )
        self.name = name
        self.types = {}
        self.exposed = {}
        self.renames = renames
        if not set(renames).isdisjoint(set(keep_names)):
            raise Exception(
                f"Names are in both `renames` and `keep_names`: {', '.join(set(renames).intersection(keep_names))}"
            )
        self.renames.update({name: name for name in keep_names})
        reserved = [name for name in self.renames.values() if name in reserved_types]
        if len(reserved) > 0:
            raise Exception(f"Cannot use reserved types: {', '.join(reserved)}")

        self.tg = TypeGraph(name="__importer__")

    def __enter__(self):
        ImporterContext.push(self)
        self.tg.__enter__()
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self.tg.__exit__(exc_type, exc_value, exc_tb)
        ImporterContext.pop()

    def add_type(self, name: str, typ: t.typedef):
        if name in self.types:
            raise Exception(f"Cannot duplicate name '{name}'")
        self.types[name] = typ

    def __call__(self, preferred_name: str, typ: t.typedef):
        return self.add_type(preferred_name, typ)

    def expose(self, name: str, fn: t.func):
        if name in self.exposed:
            raise Exception(f"Cannot expose '{name}' more than once")
        self.exposed[name] = fn

    def generate(self):
        raise Exception("Must be overriden")

    def codegen(self, cg: Codegen):
        self.generate()

        typify = Typify(self, "t")

        for header_line in self.headers:
            cg.line(header_line)
        cg.line()

        counter = itertools.count(1)
        renames = {name: f"_{self.name}_{next(counter)}_{name}" for name in self.types}
        renames.update(self.renames)

        cg.line(f"renames = {repr(renames)}")
        cg.line()

        cg.line("types = {}")

        if len(self.types) > 0:
            for name, tpe in self.types.items():
                cg.line(f"types[{repr(name)}] = {typify(tpe, name)}")

            cg.line()

        cg.line("functions = {}")
        for name, fn in self.exposed.items():
            cg.line(f"functions[{repr(name)}] = {typify(fn)}")
        cg.line()

        self.imports.add(("typegraph.importers.base.importer", "Import"))
        cg.line(
            f"return Import(importer={repr(self.name)}, renames=renames, types=types, functions=functions)"
        )

        return cg

    def imp(self, gen: bool):
        if not gen:
            return

        file = inspect.stack()[1].filename

        with open(file) as f:
            code = RedBaron(f.read())

        gen_arg = self.find_generate_arg(code)
        if not gen_arg:
            raise Exception("Expected to find generate argument import node")
        gen_arg.value = "False"

        cg = self.codegen(Codegen()).res

        for frm, imp in self.imports:
            if not code.find(
                "from_import",
                value=lambda x: x.dumps() == frm,
                targets=lambda x: x.dumps() == imp,
            ):
                code.insert(0, f"from {frm} import {imp}\n")

        name = f"import_{self.name}"
        target_node = code.find("def", lambda node: node.name == name)
        wth = code.find("with")

        comment = f"# Function generated by {type(self).__name__}. Do not change."

        if target_node is not None:
            comment_node = next(
                (
                    node
                    for node in target_node.previous_generator()
                    if node.type != "endl"
                ),
                None,
            )
            if comment_node.type != "comment":
                target_node.insert_before(comment)
            else:
                comment_node.value = comment
            target_node.value = cg
        else:
            wth.insert_before(comment)
            wth.insert_before(f"def {name}():\n{cg}")

        new_code = black.format_str(code.dumps(), mode=black.FileMode())

        with open(file, "w") as f:
            f.write(new_code)

        print(f"File updated: {file}", file=sys.stderr)
        exit(0)

    def find_generate_arg(self, code: RedBaron) -> Optional[NameNode]:
        for node in code.find_all("atomtrailers"):
            found = self.find_generate_arg_in(node)
            if found:
                return found

        return None

    def find_generate_arg_in(self, node: AtomtrailersNode) -> Optional[NameNode]:
        if len(node.node_list) != 5:
            return None

        first = node.node_list[0]

        if first.type != "name" or first.value != type(self).__name__:
            return None

        if not node.find("string", value=lambda v: literal_eval(v) == self.name):
            return None

        [imp, call] = node.node_list[-2:]
        if imp.type != "name" or imp.value != "imp":
            return None

        if call.type != "call":
            return None

        gen = call.find("name", value="True")
        if not gen:
            return None

        return gen


class ImporterContext:
    importer: Optional[Importer] = None

    @classmethod
    def push(cls, importer: Importer):
        if cls.importer is not None and cls.importer is not importer:
            raise Exception(f"Cannot replace active ImporterContext {cls.importer}")
        cls.importer = importer

    @classmethod
    def pop(cls) -> Optional[Importer]:
        temp = cls.importer
        cls.importer = None
        return temp

    @classmethod
    def get_active(cls) -> Optional[Importer]:
        return cls.importer


@frozen
class Import:
    importer: str
    renames: Dict[str, str]
    types: Dict[str, t.typedef]
    functions: Dict[str, t.func]

    def type(self, name: str):
        typ = self.types.get(name)
        if typ is None:
            raise Exception(f"Type '{name}' not found in import '{self.importer}")

    def func(self, name: str):
        fn = self.functions.get(name)
        if fn is None:
            raise Exception(f"Function '{name}' not found in import '{self.importer}'")
        return fn

    def all(self):
        return self.functions



# Copyright Metatype OÜ under the Elastic License 2.0 (ELv2). See LICENSE.md for usage.

from enum import auto
from typing import Optional

from attrs import frozen
from strenum import LowercaseStrEnum


class EffectType(LowercaseStrEnum):
    CREATE = auto()
    UPDATE = auto()
    UPSERT = auto()
    DELETE = auto()
    UNKNOWN = auto()


@frozen
class Effect:
    effect: Optional[EffectType]
    # see: https://developer.mozilla.org/en-US/docs/Glossary/Idempotent
    idempotent: bool


def none():
    return Effect(None, True)


def create(idempotent=False):
    return Effect(EffectType.CREATE, idempotent)


def update(idempotent=False):
    return Effect(EffectType.UPDATE, idempotent)


def upsert(idempotent=True):
    return Effect(EffectType.UPSERT, idempotent)


def delete(idempotent=True):
    return Effect(EffectType.DELETE, idempotent)



# Copyright Metatype OÜ under the Elastic License 2.0 (ELv2). See LICENSE.md for usage.

from typing import Optional
from typing import Union

from attrs import frozen
from typegraph import t
from typegraph.graph.typegraph import find
from typegraph.graph.typegraph import resolve_proxy
from typegraph.providers.prisma.relations import RelationshipRegister
from typegraph.providers.prisma.utils import resolve_entity_quantifier


@frozen
class TypeGenerator:
    spec: RelationshipRegister

    def get_input_type(
        self,
        tpe: t.struct,
        skip=set(),  # set of relation names, indicates related models to skip
        update=False,
        name: Optional[str] = None,
    ) -> Union[t.typedef, t.NodeProxy]:
        proxy = name and find(name)
        if proxy is not None:
            return proxy

        fields = {}
        if not isinstance(tpe, t.struct):
            raise Exception(f'expected a struct, got: "{type(tpe).__name__}"')
        for key, field_type in tpe.props.items():
            field_type = resolve_proxy(field_type)
            proxy = self.spec.proxies[tpe.name].get(key)
            relname = proxy.link_name if proxy is not None else None
            if relname is not None:
                if relname in skip:
                    continue
                relation = self.spec.relations[relname]
                nested = resolve_proxy(resolve_entity_quantifier(field_type))

                entries = {
                    "create": self.get_input_type(
                        nested, skip=skip | {relname}, name=f"Input{nested.name}Create"
                    ).optional(),
                    "connect": self.get_where_type(
                        nested, name=f"Input{nested.name}"
                    ).optional(),
                }
                if (
                    relation.side_of(tpe.name).is_left()
                    and relation.cardinality.is_one_to_many()
                ):
                    entries["createMany"] = t.struct(
                        {"data": t.array(entries["create"].of)}
                    ).optional()

                fields[key] = t.struct(entries).optional()

            elif isinstance(field_type, t.func):
                raise Exception(f'Unsupported function field "{key}"')
            else:
                if update:
                    fields[key] = field_type.optional()
                else:  # create
                    if field_type.runtime_config.get("auto", False):
                        fields[key] = field_type.optional()
                    else:
                        fields[key] = field_type.replace()  # TODO clone

        if name is None:
            return t.struct(fields)
        else:
            return t.struct(fields).named(name)

    def get_out_type(
        self, tpe: t.typedef, name: Optional[str] = None
    ) -> Union[t.typedef, t.NodeProxy]:
        proxy = name and find(name)
        if proxy is not None:
            return proxy

        if isinstance(tpe, t.func):
            return self.get_out_type(tpe.out)

        if not isinstance(tpe, t.struct):
            return tpe
        ret = t.struct({k: self.get_out_type(v) for k, v in tpe.props.items()})
        if name is None:
            return ret
        else:
            return ret.named(name)

    def get_where_type(
        self, tpe: t.struct, skip_rel=False, name: Optional[str] = None
    ) -> t.struct:
        proxy = name and find(name)
        if proxy is not None:
            return proxy

        fields = {}

        for k, v in tpe.props.items():
            # not for this runtime
            v = resolve_proxy(v)
            if v.runtime is not None and v.runtime != tpe.runtime:
                continue

            nested = resolve_proxy(resolve_entity_quantifier(v))
            # relation
            if nested.type == "object":
                if skip_rel:
                    continue
                fields[k] = self.get_where_type(nested, skip_rel=True).optional()
                continue
            if isinstance(v, t.optional):
                v = v.of
            if isinstance(v, t.NodeProxy):
                v = v.get()
            fields[k] = v.optional()

        if name is None:
            return t.struct(fields)
        else:
            return t.struct(fields).named(name)


def unsupported_cardinality(c: str):
    return Exception(f'Unsupported cardinality "{c}"')



# Copyright Metatype OÜ under the Elastic License 2.0 (ELv2). See LICENSE.md for usage.


from collections import defaultdict
from typing import Dict
from typing import Optional
from typing import TYPE_CHECKING

from attrs import frozen
from strenum import StrEnum
from typegraph import types as t
from typegraph.graph.nodes import NodeProxy
from typegraph.graph.typegraph import resolve_proxy
from typegraph.graph.typegraph import TypeGraph
from typegraph.providers.prisma.utils import resolve_entity_quantifier

if TYPE_CHECKING:
    from typegraph.providers.prisma.runtimes.prisma import PrismaRuntime


class LinkProxy(NodeProxy):
    runtime: "PrismaRuntime"
    link_name: str
    target_field: Optional[str]

    def __init__(
        self,
        g: TypeGraph,
        node: str,
        runtime: "PrismaRuntime",
        link_name: str,
        target_field: Optional[str],
    ):
        super().__init__(g, node)
        self.runtime = runtime
        self.link_name = link_name
        self.target_field = target_field


class Cardinality(StrEnum):
    ONE = "one to one"
    MANY = "one to many"

    def is_one_to_one(self):
        return self == Cardinality.ONE

    def is_one_to_many(self):
        return self == Cardinality.MANY


class Side(StrEnum):
    """
    A relationship is defined between two models:
    - the "owner", on the "left" side of the relationship, has the foreign key
    - the "ownee", on the "right" side of the relationship
    """

    LEFT = "left"
    RIGHT = "right"

    def is_left(self):
        return self == Side.LEFT

    def is_right(self):
        return self == Side.RIGHT


@frozen
class FieldRelation:
    name: str
    typ: t.struct
    field: str
    cardinality: Cardinality
    side: Side


@frozen
class Relation:
    left: FieldRelation
    right: FieldRelation

    def __attrs_post_init__(self):
        assert (
            self.left.name == self.right.name
        ), "left and right sides of a relationship must have the same name"
        assert (
            self.left.cardinality == self.right.cardinality
        ), "left and right sides of a relationship must have the same cardinality"

    @property
    def name(self):
        return self.left.name

    @property
    def left_type_name(self):
        return self.left.typ.name

    @property
    def right_type_name(self):
        return self.right.typ.name

    def side_of(self, type_name: str) -> Optional[Side]:
        if self.left_type_name == type_name:
            return Side.LEFT
        if self.right_type_name == type_name:
            return Side.RIGHT
        return None

    @property
    def cardinality(self):
        return self.right.cardinality


class RelationshipRegister:
    """
    Relationships are defined by `LinkProxy` NodeProxy types on the model types (`t.struct`).
    """

    runtime: "PrismaRuntime"
    types: Dict[str, t.struct]
    relations: Dict[str, Relation]
    proxies: Dict[str, Dict[str, LinkProxy]]

    def __init__(self, runtime):
        self.runtime = runtime
        self.types = {}
        self.relations = {}
        self.proxies = defaultdict(dict)

    def get_left_proxy(self, left_field: str, right: LinkProxy) -> LinkProxy:
        right_type = resolve_proxy(resolve_entity_quantifier(right.get()))
        prop_type = right_type.props[right.target_field]
        if not isinstance(prop_type, LinkProxy):
            prop_type = self.proxies[right_type.name][right.target_field]
        left = prop_type
        assert left.link_name == right.link_name
        left_type = left.get()
        assert left_type.type == "object", f"Expected object, got {left_type.type}"

        if left.target_field is not None:
            assert left.target_field == left_field
        else:
            left.target_field = left_field

        return left

    def get_right_proxy(self, right_field: str, left: LinkProxy) -> LinkProxy:
        left_type = left.get()
        prop_type = left_type.props[left.target_field]
        if not isinstance(prop_type, LinkProxy):
            prop_type = self.proxies[left_type.name][left.target_field]
        right = prop_type
        assert left.link_name == right.link_name
        assert right.get().type in [
            "optional",
            "array",
        ], f"Expected optional or array, got '{right.get().type}'"
        if right.target_field is not None:
            assert right.target_field == right_field
        else:
            right.target_field = right_field

        return right

    def get_link_from_proxy(
        self, proxy: LinkProxy, target_type: t.struct, target_field: str
    ):
        typ = proxy.get()

        if typ.type in ["optional", "array"]:
            right = proxy
            if right.target_field is None:
                right_type = resolve_proxy(resolve_entity_quantifier(right.get()))
                # TODO
                # match by link name
                fields = [
                    f
                    for f, ty in right_type.props.items()
                    if isinstance(ty, LinkProxy) and ty.link_name == right.link_name
                ]
                if len(fields) == 0:
                    print(
                        f"proxies: {repr({k: f'{ln.link_name}, {ln.name}, {ln.target_field}' for k, ln in self.proxies[right_type.name].items()})}"
                    )
                    fields = [
                        f
                        for f, p in self.proxies[right_type.name].items()
                        if p.link_name == right.link_name
                    ]
                print(f"right: link name={right.link_name}")
                assert len(fields) == 1, f"got: {repr(fields)}"
                right.target_field = fields[0]

            left_field = target_field
            left = self.get_left_proxy(left_field, right)
            return left, right

        assert typ.type == "object"
        left = proxy
        if left.target_field is None:
            left_type = left.get()
            fields = [
                f
                for f, ty in left_type.props.items()
                if isinstance(ty, LinkProxy) and ty.link_name == left.link_name
            ]
            assert len(fields) == 1
            left.target_field = fields[0]

        right_field = target_field
        right = self.get_right_proxy(right_field, left)
        return left, right

    def fill_proxy_target_field(self, proxy: LinkProxy, model: t.struct):
        if proxy.target_field is not None:
            return
        typ = proxy.get()

        if isinstance(typ, t.struct):
            left_model = typ

            fields = [
                f
                for f, ty in left_model.props.items()
                if isinstance(ty, LinkProxy) and ty.link_name == proxy.link_name
            ]
            assert (
                len(fields) <= 1
            ), f"Relationship names must be unique: got multiple '{proxy.link_name}' on '{left_model.name}': {', '.join(fields)}"
            if len(fields) == 1:
                proxy.target_field = fields[0]
                return

            fields = [
                f
                for f, ty in left_model.props.items()
                if isinstance(ty, t.optional)
                or isinstance(ty, t.array)
                and ty.of.name == model.name
            ]
            assert (
                len(fields) <= 1
            ), f"Ambiguous target field for '{proxy.link_name}' on '{left_model.name}': {' or '.join(fields)} ?"
            assert (
                len(fields) == 1
            ), f"No target for '{proxy.link_name}' on '{left_model.name}'"
            proxy.target_field = fields[0]
            return

        assert isinstance(typ, t.optional) or isinstance(typ, t.array)
        right_model = resolve_proxy(typ.of)
        assert isinstance(right_model, t.struct)

        fields = [
            f
            for f, ty in right_model.props.items()
            if isinstance(ty, LinkProxy) and ty.link_name == proxy.link_name
        ]
        assert (
            len(fields) <= 1
        ), f"Relationship names must be unique: got multiple '{proxy.link_name}' on '{right_model.name}': {', '.join(fields)}"
        if len(fields) == 1:
            proxy.target_field = fields[0]
            return

        fields = [f for f, ty in right_model.props.items() if ty.name == model.name]
        assert (
            len(fields) <= 1
        ), f"Ambiguous target field for '{proxy.link_name}' on '{right_model.name}': {' or '.join(fields)} ?"
        assert (
            len(fields) == 1
        ), f"No target for '{proxy.link_name}' on '{right_model.name}'"
        proxy.target_field = fields[0]

    # ensure that there is a proxy on the other side
    # TODO: full validation??
    def ensure_proxy_target(self, proxy: LinkProxy, field: str, model: t.struct):
        typ = proxy.get()
        self.fill_proxy_target_field(proxy, model)

        if isinstance(typ, t.struct):
            # TODO: find target field
            left_model = typ
            # assert proxy.target_field is not None
            right_proxy = left_model.props.get(proxy.target_field)

            assert (
                right_proxy is not None
            ), f"Expected field '{proxy.target_field}' on '{left_model.name}'"

            if isinstance(right_proxy, LinkProxy):
                assert right_proxy.link_name == proxy.link_name
                if right_proxy.target_field is None:
                    right_proxy.target_field = field
                else:
                    assert right_proxy.target_field == field
                registered_proxy = self.proxies[typ.name][proxy.target_field]
                if registered_proxy is None:
                    self.proxies[typ.name][proxy.target_field] = right_proxy
                else:
                    assert registered_proxy == right_proxy

            else:
                # wrap type in a proxy
                self.proxies[left_model.name][proxy.target_field] = self.runtime.link(
                    right_proxy, proxy.link_name, field
                )

        else:
            assert isinstance(typ, t.optional) or isinstance(
                typ, t.array
            ), "LinkProxy can only wrap a struct, optional or array"
            right_model = resolve_proxy(typ.of)
            assert isinstance(
                right_model, t.struct
            ), f"Expected an struct for a model, got '{right_model.type}'"

            left_proxy = right_model.props.get(proxy.target_field)
            assert (
                left_proxy is not None
            ), f"Expected field '{proxy.target_field}' on '{right_model.name}'"

            if isinstance(left_proxy, LinkProxy):
                assert left_proxy.link_name == proxy.link_name
                if left_proxy.target_field is None:
                    left_proxy.target_field = field
                else:
                    assert left_proxy.target_field == field

                registered_proxy = self.proxies[right_model.name].get(
                    proxy.target_field, None
                )
                if registered_proxy is None:
                    self.proxies[right_model.name][proxy.target_field] = left_proxy
                else:
                    assert registered_proxy == left_proxy

            else:
                # wrap type in a proxy
                self.proxies[right_model.name][proxy.target_field] = self.runtime.link(
                    left_proxy, proxy.link_name, field
                )

    def register_type(self, typ: t.struct):
        self.types[typ.name] = typ
        # TODO set runtime, propagate runtime
        proxies = self.proxies[typ.name]
        for prop_name, prop_type in typ.props.items():
            if isinstance(prop_type, LinkProxy):
                proxies[prop_name] = prop_type
                self.ensure_proxy_target(prop_type, prop_name, typ)
                continue
            if prop_name in proxies:
                continue

            prop_type = resolve_proxy(prop_type)
            if isinstance(prop_type, t.struct):
                # find all props of `prop_type` that define a relationship to `typ` (RIGHT proxy)
                fields = [
                    f
                    for f, ty in prop_type.props.items()
                    if isinstance(ty, LinkProxy)
                    and resolve_entity_quantifier(ty.get()).name == typ.name
                ]

                # todo: assert link_name
                assert (
                    len(fields) <= 1
                ), f"Link must be explicitly specified when there are more than one links targetting the same model: on '{prop_type.name}': {' or '.join(fields)}"

                if len(fields) == 1:
                    link_name = prop_type.props[fields[0]].link_name
                    proxies[prop_name] = self.runtime.link(
                        prop_type, link_name, fields[0]
                    )
                    continue

                # len(fields) == 0
                # find all props of `prop_type` of type optional/array of `typ`
                fields = [
                    f
                    for f, ty in prop_type.props.items()
                    if (isinstance(ty, t.optional) or isinstance(ty, t.array))
                    and resolve_entity_quantifier(ty).name == typ.name
                ]

                assert (
                    len(fields) <= 1
                ), f"Link must be explicitly specified when there are more than one implicit links targetting the same model: on '{prop_type.name}'"
                assert (
                    len(fields) == 1
                ), f"No field match to a relationship to '{typ.name}' on '{prop_type.name}'"
                # TODO handle eventual name clash
                link_name = f"_{typ.name}_to_{prop_type.name}"
                proxies[prop_name] = self.runtime.link(prop_type, link_name, fields[0])
                self.proxies[prop_type.name][fields[0]] = self.runtime.link(
                    prop_type.props[fields[0]], link_name, prop_name
                )
                self.ensure_proxy_target(proxies[prop_name], prop_name, typ)
                continue

            if isinstance(prop_type, t.optional) or isinstance(prop_type, t.array):
                model = resolve_proxy(prop_type.of)
                if not isinstance(model, t.struct):
                    continue

                # find all props of `model` that define a relationship to `typ` (LEFT proxy)
                fields = [
                    f
                    for f, ty in model.props.items()
                    if isinstance(ty, LinkProxy) and ty.name == typ.name
                ]

                assert (
                    len(fields) <= 1
                ), f"Ambiguous link targetting '{typ.name}' on '{model.name}': {' or '.join(fields)}?"

                if len(fields) == 1:
                    link_name = model.props[fields[0]].link_name
                    proxies[prop_name] = self.runtime.link(
                        prop_type, link_name, fields[0]
                    )
                    continue

                # len(fields) == 0
                # find all props of `model` of `typ`
                fields = [
                    f
                    for f, ty in model.props.items()
                    if ty.name == typ.name
                    # if isinstance(resolve_proxy(ty))
                ]
                assert len(fields) <= 1
                assert len(fields) == 1
                # TODO handle eventual name clash
                # generate link_name
                link_name = f"_{model.name}_to_{typ.name}"
                proxies[prop_name] = self.runtime.link(prop_type, link_name, fields[0])
                self.proxies[model.name][fields[0]] = self.runtime.link(
                    model.props[fields[0]], link_name, prop_name
                )

    def manage(self, typ: t.struct):
        if typ.name in self.types:
            return

        assert typ.type == "object", f"Prisma cannot manage '{typ.type}' types"
        self.register_type(typ)

        deps = []

        for prop_name, proxy in self.proxies[typ.name].items():
            left_proxy, right_proxy = self.get_link_from_proxy(proxy, typ, prop_name)

            rel = create_relationship(left_proxy, right_proxy)
            self.relations[rel.name] = rel
            deps.append(rel.left.typ)
            deps.append(rel.right.typ)

        for ty in deps:
            self.manage(ty)


def create_relationship(left_proxy: LinkProxy, right_proxy: LinkProxy) -> Relation:
    left_type = left_proxy.get()
    left_model = left_type
    right_type = right_proxy.get()
    right_model = resolve_proxy(resolve_entity_quantifier(right_type))
    assert left_model.type == "object", "Relationships can only be defined on objects"
    assert (
        right_type.type in ["optional", "array"] and right_model.type == "object"
    ), "Right side of a relationship must be an array or optional"

    left_field = left_proxy.target_field
    right_field = right_proxy.target_field
    assert left_model.props[left_field].name == right_type.name
    assert right_model.props[right_field].name == left_type.name

    cardinality, q = (
        (Cardinality.ONE, "?")
        if right_type.type == "optional"
        else (Cardinality.MANY, "[]")
    )
    relname = left_proxy.link_name

    return Relation(
        left=FieldRelation(
            name=relname,
            typ=left_model,
            field=right_proxy.target_field,  # what??
            cardinality=cardinality,
            side=Side.LEFT,
        ),
        right=FieldRelation(
            name=relname,
            typ=right_model,
            field=left_proxy.target_field,  # what??
            cardinality=cardinality,
            side=Side.RIGHT,
        ),
    )



# Copyright Metatype OÜ under the Elastic License 2.0 (ELv2). See LICENSE.md for usage.

from typing import Dict, List, Literal, Optional

from attrs import define, field, frozen


@define
class Auth:
    """
    Authentication options
    """

    name: str
    protocol: str
    auth_data: Dict[str, str]

    @classmethod
    def oauth2(
        cls,
        name: str,
        authorize_url: str,
        access_url: str,
        scopes: str,
        profile_url: Optional[str] = None,
    ) -> "Auth":
        """
        OAuth2 authentication
        """
        return Auth(
            name,
            "oauth2",
            dict(
                authorize_url=authorize_url,
                access_url=access_url,
                scopes=scopes,
                profile_url=profile_url,
            ),
        )

    # deno eval 'await crypto.subtle.generateKey({name: "ECDSA", namedCurve: "P-384"}, true, ["sign", "verify"]).then(k => crypto.subtle.exportKey("jwk", k.publicKey)).then(JSON.stringify).then(console.log);'
    @classmethod
    def jwk(cls, name: str, args=None) -> "Auth":
        return Auth(name, "jwk", args if args is not None else {})

    @classmethod
    def basic(cls, users: List[str]) -> "Auth":
        return Auth("basic", "basic", {"users": users})


@define
class Cors:
    """
    CORS options
    """

    allow_origin: List[str] = field(factory=list)
    allow_headers: List[str] = field(factory=list)
    expose_headers: List[str] = field(factory=list)
    allow_credentials: bool = True
    max_age: Optional[int] = None


@define
class Rate:
    """
    Rate limiting options
    """

    window_limit: int
    window_sec: int
    query_limit: int
    context_identifier: Optional[str] = None
    local_excess: int = 0


@frozen
class Code:
    name: str
    source: str
    type: Literal["func", "module"] = field(default="func")



# Copyright Metatype OÜ under the Elastic License 2.0 (ELv2). See LICENSE.md for usage.

from typing import List, Optional, Union

from attrs import field, frozen

from typegraph import effects
from typegraph import types as t
from typegraph.effects import Effect
from typegraph.graph.builder import Collector
from typegraph.graph.nodes import Node, NodeProxy
from typegraph.graph.typegraph import TypegraphContext
from typegraph.providers.prisma.relations import LinkProxy
from typegraph.providers.prisma.schema import RelationshipRegister, build_model
from typegraph.providers.prisma.type_generator import TypeGenerator
from typegraph.runtimes.base import Materializer, Runtime
from typegraph.utils.attrs import SKIP, always, required


def comp_exp(tpe):
    name = f"bool_exp_{tpe.name}"
    return t.struct(
        {
            "_eq": t.proxy(name),
            "_gt": t.proxy(name),
            "_gte": t.proxy(name),
            "_in": t.array(t.proxy(name)),
            "_is_null": t.boolean(),
            "_lt": t.proxy(name),
            "_lte": t.proxy(name),
            "_neq": t.proxy(name),
            "_nin": t.array(t.proxy(name)),
        }
    ).named(name)


def bool_exp(tpe: t.struct):
    name = f"bool_exp_{tpe.name}"
    g = TypegraphContext.get_active()
    return t.struct(
        {
            "_and": t.array(g(name)),
            "_not": g(name),
            "_or": t.array(g(name)),
        }
    ).named(name)


def sql_select(tpe: t.struct):
    cols = tpe.props.keys()

    return t.struct(
        {
            "distinct_on": t.array(t.enum(cols)).named(f"sql_distinct_on_{tpe.name}"),
            "limit": t.unsigned_integer(),
            "offset": t.unsigned_integer(),
            "order_by": t.array(t.tuple([t.enum(cols), t.string()])).named(
                f"sql_order_by_{tpe.name}"
            ),
            "where": bool_exp(tpe).optional(),
        }
    )


def sql_insert(tpe: t.struct):
    return t.struct(
        {
            "objects": t.array(
                t.struct(
                    {
                        field_name: field_type
                        for field_name, field_type in tpe.props.items()
                    }
                )
            ),
        }
    )


def sql_update(tpe: t.struct):
    return t.struct(
        {
            "_set": t.struct(
                {field_name: field_type for field_name, field_type in tpe.props.items()}
            ),
            "where": bool_exp(tpe),
        }
    )


def sql_delete(tpe: t.struct):
    return t.struct({"where": bool_exp(tpe)})


@frozen
class PrismaOperationMat(Materializer):
    runtime: "PrismaRuntime"
    table: str
    operation: str
    materializer_name: str = always("prisma_operation")
    effect: Effect = required()


# https://github.com/prisma/prisma-engines/tree/main/query-engine/connector-test-kit-rs/query-engine-tests/tests/queries
@frozen
class PrismaRuntime(Runtime):
    """
    > A database ORM runtime.

    Attributes:
        `name`: Name of prisma runtime
        `connection_string_secret`: Name of the secret that contains the connection string
        that will be used to connect to the database

    Example:
    ```python
    with TypeGraph("prisma-runtime-example") as g:
        db = PrismaRuntime("main_db", "DB_CONNECTION")

        user = t.struct(
            {
                "id": t.uuid().config("id", "auto"),
                "email": t.email(),
            }
        )

        g.expose(
            createUser=db.create(user).add_policy(public)
        )
    ```

    ### Models

    Any `t.struct` that is passed to a generator of a `PrismaRuntime`
    defines a model.
    Models must have an ID field specified by the `"id"` config.

    Here is the list of all the available configs for model fields:

    | Config | Effect |
    |---|---|
    | `id` | defines the field ID for the model (a.k.a. primary key) |
    | `auto` | the value of this field can be auto generated; supported for `t.integer()` (auto-increment) and `t.uuid()` |
    | `unique` | make this field unique among all instances of the model |

    ### Relationships

    The `PrismaRuntime` supports two kinds of relationship between models.

    | Relationship | Field type in Model1 | Field type in Model2 |
    |---|---|---|
    |One to one| `g("Model2")` | `g("Model1").optional()` |
    |One to many| `g("Model2")` | `t.array(g("Model1"))` |

    Relationship fields must be defined on both sides of the relationship.
    A relationship is always defined for `t.struct` types and `t.optional` or
    `t.array` of `t.struct`s.

    Relatioships can also be defined implicitly using the `link` instance method
    of `PrismaRuntime`.

    Example:

    ```python
    runtime = PrismaRuntime("example", "POSTGRES")

    user = t.struct(
        {
            "id": t.uuid().config("id", "auto"),
            "email": t.email().config("unique"),
            "posts": t.array(g("Post")),
        }
    ).named("User")

    post = t.struct(
        {
            "id": t.uuid().config("id", "auto"),
            "title": t.string(),
            "author": g("User"),
        }
    ).named("Post")
    ```

    ### Generators

    Generators are instance methods of `PrismaRuntime` that can be used
    to generate a `t.func` that represents a specific operation on a specific
    model of the runtime.
    They match to the model queries defined for the
    [prisma client API](https://www.prisma.io/docs/reference/api-reference/prisma-client-reference).
    for the type of the input `t.struct` and the return type.

    Example:

    ```python
    with TypeGraph("prisma-runtime-example") as g:
        db = PrismaRuntime("main_db", "DB_CONNECTION")

        user = t.struct(
            {
                "id": t.uuid().config("id", "auto"),
                "email": t.email(),
            }
        )

        g.expose(
            createUser=db.create(user).add_policy(public),
            findUser=db.find_unique(user).add_policy(public),
            findManyUsers=db.find_many(user).add_policy(public),
        )
    ```

    Here is a list of all available generators:
    - `find_unique`
    - `find_many`
    - `create`
    - `update`
    - `delete`
    - `delete_many`


    """

    name: str
    connection_string_secret: str
    runtime_name: str = always("prisma")
    spec: RelationshipRegister = field(init=False, hash=False, metadata={SKIP: True})

    def __attrs_post_init__(self):
        object.__setattr__(self, "spec", RelationshipRegister(self))

    def link(
        self, typ: Union[t.TypeNode, str], name: str, field: Optional[str] = None
    ) -> t.TypeNode:
        """
        Explicitly declare a relationship between models. The return value of
        this function shall be the type of a property of a `t.struct` that
        defines a model.
        If the other end of the relationship is also defined using `link`,
        both links must have the same name.

        Arguments:
            name:     name of the relationship
            field:    name of the target field on the target model

        Example:
        ```python
        runtime = PrismaRuntime("example", "POSTGRES")

        user = t.struct(
            {
                "id": t.uuid().config("id", "auto"),
                "email": t.email().config("unique"),
                "posts": runtime.link(t.array(g("Post")), "postAuthor"),
            }
        ).named("User")

        post = t.struct(
            {
                "id": t.uuid().config("id", "auto"),
                "title": t.string(),
                "author": runtime.link(g("User"), "postAuthor"),
            }
        ).named("Post")
        ```
        """
        if isinstance(typ, t.typedef) or isinstance(typ, NodeProxy):
            g = typ.graph
            if isinstance(typ, t.typedef):
                typ.register_name()
            typ = typ.name
        else:
            g = TypegraphContext.get_active()
        return LinkProxy(g, typ, self, name, field)

    @property
    def __typegen(self):
        return TypeGenerator(spec=self.spec)

    def queryRaw(self, query: str, *, effect: Effect) -> t.func:
        """Generate a raw SQL query operation"""
        return t.func(
            t.struct(
                {
                    "parameters": t.json(),
                }
            ).named("QueryRawInp"),
            t.array(t.json()),
            PrismaOperationMat(self, query, "queryRaw", effect=effect),
        )

    def executeRaw(self, query: str, *, effect: Effect) -> t.func:
        """Generate a raw SQL query operation without return"""
        return t.func(
            t.struct(
                {
                    "parameters": t.json().optional().default("[]"),
                }
            ).named(f"ExecuteRawInp_{TypegraphContext.get_active().next_type_id()}"),
            t.integer(),
            PrismaOperationMat(self, query, "executeRaw", effect=effect),
        )

    def find_unique(self, tpe: Union[t.struct, t.NodeProxy]) -> t.func:
        self.__manage(tpe)
        typegen = self.__typegen
        return t.func(
            t.struct(
                {"where": typegen.get_where_type(tpe).named(f"{tpe.name}WhereUnique")}
            ),
            typegen.get_out_type(tpe).named(f"{tpe.name}UniqueOutput").optional(),
            PrismaOperationMat(self, tpe.name, "findUnique", effect=effects.none()),
        )

    def find_many(self, tpe: Union[t.struct, t.NodeProxy]) -> t.func:
        self.__manage(tpe)
        typegen = self.__typegen
        return t.func(
            t.struct(
                {
                    "where": typegen.get_where_type(tpe)
                    .named(f"{tpe.name}Where")
                    .optional()
                }
            ),
            t.array(typegen.get_out_type(tpe).named(f"{tpe.name}Output")),
            PrismaOperationMat(self, tpe.name, "findMany", effect=effects.none()),
        )

    def create(self, tpe: Union[t.struct, t.NodeProxy]) -> t.func:
        self.__manage(tpe)
        typegen = self.__typegen
        return t.func(
            t.struct(
                {
                    "data": typegen.get_input_type(tpe).named(f"{tpe.name}CreateInput"),
                }
            ),
            typegen.get_out_type(tpe).named(f"{tpe.name}CreateOutput"),
            PrismaOperationMat(self, tpe.name, "createOne", effect=effects.create()),
        )

    def update(self, tpe: Union[t.struct, t.NodeProxy]) -> t.func:
        self.__manage(tpe)
        typegen = self.__typegen
        return t.func(
            t.struct(
                {
                    "data": typegen.get_input_type(tpe, update=True).named(
                        f"{tpe.name}UpdateInput"
                    ),
                    "where": typegen.get_where_type(tpe).named(
                        f"{tpe.name}UpdateOneWhere"
                    ),
                }
            ),
            typegen.get_out_type(tpe).named(f"{tpe.name}UpdateOutput"),
            PrismaOperationMat(
                self, tpe.name, "updateOne", effect=effects.update(True)
            ),
        )

    def delete(self, tpe: Union[t.struct, t.NodeProxy]) -> t.func:
        self.__manage(tpe)
        typegen = self.__typegen
        return t.func(
            t.struct(
                {"where": typegen.get_where_type(tpe).named(f"{tpe.name}DeleteInput")},
            ),
            typegen.get_out_type(tpe).named(f"{tpe.name}DeleteOutput"),
            PrismaOperationMat(self, tpe.name, "deleteOne", effect=effects.delete()),
        )

    def delete_many(self, tpe: Union[t.struct, t.NodeProxy]) -> t.func:
        self.__manage(tpe)
        typegen = self.__typegen
        return t.func(
            t.struct(
                {
                    "where": typegen.get_where_type(tpe).named(
                        f"{tpe.name}DeleteManyWhereInput"
                    ),
                }
            ),
            t.struct({"count": t.integer()}).named(f"{tpe.name}BatchDeletePayload"),
            PrismaOperationMat(self, tpe.name, "deleteMany", effect=effects.delete()),
        )

    def __manage(self, tpe):
        tpe._propagate_runtime(self)
        self.spec.manage(tpe)

    def __datamodel(self):
        models = [build_model(ty, self.spec) for ty in self.spec.types]
        return "\n\n".join(models)
        # return PrismaSchema(self.managed_types.values()).build()

    def data(self, collector: Collector) -> dict:
        data = super().data(collector)
        data["data"].update(
            datamodel=self.__datamodel(),
            connection_string_secret=self.connection_string_secret,
            models=[collector.index(tp) for tp in self.spec.types.values()],
        )
        return data

    @property
    def edges(self) -> List[Node]:
        return super().edges + list(self.spec.types.values())

    def insert_one(self, tpe):
        return self.gen_create(tpe)


@frozen
class PrismaMigrationRuntime(Runtime):
    runtime_name: str = always("prisma_migration")


@frozen
class PrismaApplyMat(Materializer):
    runtime: Runtime = PrismaMigrationRuntime()
    materializer_name: str = always("apply")
    effect: Effect = always(effects.upsert())


@frozen
class PrismaDeployMat(Materializer):
    runtime: Runtime = PrismaMigrationRuntime()
    materializer_name: str = always("deploy")
    effect: Effect = always(effects.upsert())


@frozen
class PrismaCreateMat(Materializer):
    runtime: Runtime = PrismaMigrationRuntime()
    materializer_name: str = always("create")
    effect: Effect = always(effects.create())


@frozen
class PrismaDiffMat(Materializer):
    runtime: Runtime = PrismaMigrationRuntime()
    materializer_name: str = always("diff")
    effect: Effect = always(effects.none())



# Copyright Metatype OÜ under the Elastic License 2.0 (ELv2). See LICENSE.md for usage.

import attrs
from attrs import Attribute, field

SKIP = "asdict_skip"


def always(value):
    return field(default=value, init=False)


def required():
    return field(init=True, kw_only=True)


def asdict(inst):
    # We do not use the `_value` parameter, but it is part of the signature
    # of the filter used by `attrs.asdict`
    def filter(attr: Attribute, _value):
        if attr.metadata is not None and attr.metadata.get(SKIP, False):
            return False
        return True

    return attrs.asdict(inst, filter=filter, recurse=False)



# Copyright Metatype OÜ under the Elastic License 2.0 (ELv2). See LICENSE.md for usage.

from typing import Optional, Tuple

from attrs import frozen

from typegraph import effects
from typegraph import types as t
from typegraph.effects import Effect
from typegraph.runtimes.base import Materializer, Runtime
from typegraph.utils.attrs import always, required


@frozen
class GraphQLRuntime(Runtime):
    endpoint: str
    runtime_name: str = always("graphql")

    def query(
        self, inp: t.struct, out: t.typedef, *, path: Optional[Tuple[str, ...]] = None
    ):
        if path is not None and len(path) == 0:
            raise Exception("Unexpected empty path")
        return t.func(inp, out, QueryMat(self, path))

    def mutation(
        self,
        inp: t.struct,
        out: t.typedef,
        path: Optional[Tuple[str, ...]] = None,
        **kwargs
    ):
        if path is not None and len(path) == 0:
            raise Exception("Unexpected empty path")
        return t.func(inp, out, MutationMat(self, path, **kwargs))


@frozen
class QueryMat(Materializer):
    runtime: Runtime
    path: Optional[Tuple[str]]
    materializer_name: str = always("query")
    effect: Effect = always(effects.none())


@frozen
class MutationMat(Materializer):
    runtime: Runtime
    path: Optional[Tuple[str]]
    materializer_name: str = always("mutation")
    effect: Effect = required()



# Copyright Metatype OÜ under the Elastic License 2.0 (ELv2). See LICENSE.md for usage.

import sys
from typing import Dict


def merge(*dicts: Dict):
    ret = dict()
    for d in dicts:
        ret.update(d)
    return ret


def drop_nones(d: Dict, **kwargs) -> Dict:
    return {k: v for k, v in merge(d, kwargs).items() if v is not None}


def pick(d: Dict, *largs):
    return {k: d.get(k) for k in largs}


def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)



# Copyright Metatype OÜ under the Elastic License 2.0 (ELv2). See LICENSE.md for usage.

import json
from typing import Literal, Optional

import httpx
from box import Box
from graphql import get_introspection_query

from typegraph import t
from typegraph.importers.base.importer import Importer
from typegraph.importers.base.typify import TypifyMat

# map type kind to type key of fields
OBJECT_TYPES = {
    "OBJECT": "fields",
    "INTERFACE": "fields",
    "INPUT_OBJECT": "inputFields",
}

SCALAR_TYPES = {
    "Int": lambda: t.integer(),
    "Long": lambda: t.integer(),
    "Float": lambda: t.number(),
    "String": lambda: t.string(),
    "Boolean": lambda: t.boolean(),
    "ID": lambda: t.string(),
    "Char": lambda: t.string(),
}


class GraphQLImporter(Importer):
    intros: Box

    def __init__(self, name: str, url: str, *, file: Optional[str] = None):
        super().__init__(name)
        self.imports.add(("typegraph.runtimes.graphql", "GraphQLRuntime"))
        self.headers.append(f"{name}=GraphQLRuntime('{url}')")

        if file is None:
            res = httpx.post(url, json={"query": get_introspection_query()})
            self.intros = Box(res.json()).data
        else:
            with open(file) as f:
                self.intros = Box(json.loads(f.read()))

    def generate(self):
        schema = self.intros["__schema"]
        queryType = schema.queryType.name if schema.queryType is not None else None
        mutationType = (
            schema.mutationType.name if schema.mutationType is not None else None
        )

        queries, mutations = None, None

        for tpe in schema.types:
            if tpe.kind == "SCALAR" or tpe.name.startswith("__"):
                continue
            if tpe.name == queryType:
                queries = tpe
                continue
            if tpe.name == mutationType:
                mutations = tpe
                continue
            self.add_type_from_node(tpe)

        def expose(f, method: Literal["query", "mutation"]):
            with self:
                self.expose(
                    f.name,
                    t.func(
                        t.struct(
                            {
                                arg.name: self.type_from_node(arg.type)
                                for arg in f.args or []
                            }
                        ),
                        self.type_from_node(f.type),
                        TypifyMat(
                            lambda inp, out: f"{self.name}.{method}({inp}, {out})"
                        ),
                    ),
                )

        for q in queries.fields if queries is not None else []:
            expose(q, "query")
        for m in mutations.fields if mutations is not None else []:
            expose(m, "mutation")

    def add_type_from_node(self, node: Box):
        with self as imp:
            imp(node.name, self.non_optional_type_from_node(node))

    def non_optional_type_from_node(self, node: Box):
        if node.kind == "SCALAR":
            return self.scalar_type(node.name)

        if node.kind == "ENUM":
            if not hasattr(node, "enumValues"):
                return t.proxy(node.name)
            return t.string().enum([val.name for val in node.enumValues])

        if node.kind == "LIST":
            return t.array(self.type_from_node(node.ofType))

        if node.kind == "UNION":
            return t.union([self.type_from_node(t) for t in node.possibleTypes])

        if node.kind in OBJECT_TYPES:
            if not hasattr(node, OBJECT_TYPES[node.kind]):
                return t.proxy(node.name)

            return t.struct(
                {
                    field.name: self.type_from_node(field.type)
                    for field in node[OBJECT_TYPES[node.kind]]
                }
            )

        raise Exception(f"Unsupported type kind {node.kind}")

    def type_from_node(self, node: Box) -> t.typedef:
        if node.kind == "NON_NULL":
            return self.non_optional_type_from_node(node.ofType)
        else:
            return t.optional(self.non_optional_type_from_node(node))

    def scalar_type(self, name: str) -> t.typedef:
        return SCALAR_TYPES[name]()



# Copyright Metatype OÜ under the Elastic License 2.0 (ELv2). See LICENSE.md for usage.

import importlib
import json
import pkgutil
from argparse import ArgumentParser
from pathlib import Path
from typing import List

import attrs
from attrs import define
from frozendict import frozendict

from typegraph.graph.typegraph import TypeGraph


# TDM is: typegraph definition module, defining one or more typegraphs
@define
class LoadedTdm:
    path: str
    typegraphs: List[dict]


@define
class TypegraphError:
    path: str
    message: str


def import_file(path: str) -> List[TypeGraph]:
    # Import typegraphs from a TDM.
    path = Path(path)

    spec = importlib.util.spec_from_file_location(path.name, path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    return find_typegraphs(module)


def import_modules(module, recursive=True):
    modules = {}

    for loader, parent_name, is_pkg in pkgutil.walk_packages(module.__path__):
        name = f"{module.__name__}.{parent_name}"
        modules[name] = importlib.import_module(name)

        if recursive and is_pkg:
            modules.update(import_modules(name))

    typegraphs = []
    for mod in modules.values():
        for tg in find_typegraphs(mod):
            typegraphs.append(tg)

    return typegraphs


def find_typegraphs(module) -> List[TypeGraph]:
    ret = []
    for attr in dir(module):
        obj = getattr(module, attr)
        if isinstance(obj, TypeGraph):
            ret.append(obj)
    return ret


def cmd():
    parser = ArgumentParser()
    parser.add_argument("module")
    parser.add_argument("--pretty", action="store_true")
    # TODO option: output file

    args = parser.parse_args()

    tgs = import_file(args.module)

    class JSONEncoder(json.JSONEncoder):
        def default(self, o):
            if attrs.has(o):
                return attrs.asdict(o)
            if isinstance(o, frozendict):
                return dict(o)
            return super().default(o)

    opt = dict(indent=2) if args.pretty else {}
    output = json.dumps([tg.build() for tg in tgs], cls=JSONEncoder, **opt)
    print(output)



# Copyright Metatype OÜ under the Elastic License 2.0 (ELv2). See LICENSE.md for usage.

import json
import pathlib
import re
from sys import stderr
from typing import Callable, Dict, List, Optional, Tuple
from urllib.parse import urljoin

import httpx
import semver
import yaml
from box import Box

from typegraph import t
from typegraph.importers.base.importer import Importer
from typegraph.importers.base.typify import TypifyMat
from typegraph.utils.jsonschema import TypedefFromJsonSchema

MIME_TYPES = Box(
    {
        "json": "application/json",
        "urlenc": "application/x-www-form-urlencoded",
        "multipart": "multipart/form-data",
    }
)


def ref_to_name(ref: str) -> str:
    match = re.match(r"^#/components/schemas/(\w+)$", ref)
    if not match:
        raise Exception(f"Could not resolve $ref '{ref}'")
    return match.group(1)


class OpenApiImporter(Importer):
    base_url: str
    specification: Box
    additional_types: List[str]
    typedef_from_jsonschema: TypedefFromJsonSchema

    def __init__(
        self,
        name: str,
        *,
        url: Optional[str] = None,
        file: Optional[str] = None,
        base_url: Optional[str] = None,
        renames: Dict[str, str] = {},
        keep_names: List[str] = [],
    ):
        """Requires either only `url` or `file` and `base_url`"""

        super().__init__(name, renames=renames, keep_names=keep_names)

        self.name = name
        self.additiona_types = []

        if url is not None:
            assert file is None and base_url is None

            res = httpx.get(url)
            if res.headers.get("content-type").find("yaml") >= 0 or re.search(
                r"\.yaml$", url
            ):
                self.specification = Box(yaml.safe_load(res.text))
            else:  # suppose it is JSON
                self.specification = Box(res.json())

            self.base_url = url

        else:
            assert file is not None and base_url is not None
            self.base_url = base_url

            path = pathlib.Path(file)
            with open(path) as f:
                if path.suffix == ".json":
                    self.specification = Box(json.loads(f.read()))
                elif path.suffix in [".yaml", ".yml"]:
                    self.specification = Box(yaml.safe_load(f.read()))
                else:
                    raise Exception(
                        f"Expected a JSON or a YAML file, but got a file with extension '{path.suffix}'"
                    )

        self.typedef_from_jsonschema = TypedefFromJsonSchema(
            ref_to_name, lambda ref: self.resolve_ref(ref)
        )

        self.imports.add(("typegraph.runtimes.http", "HTTPRuntime"))

        if "servers" in self.specification:
            server_url = self.specification.servers[0].url
        else:
            server_url = "/"
        url = urljoin(self.base_url, server_url)
        self.headers.append(f"{name} = HTTPRuntime({repr(url)})")

    def resolve_ref(self, obj: Box):
        if "$ref" not in obj:
            return obj

        match = re.match(r"#/components/([^/]+)/([^/]+)$", obj["$ref"])
        if not match:
            raise Exception(f'Unsupported (external?) reference "{obj["$ref"]}"')
        return self.specification.components[match.group(1)][match.group(2)]

    def generate(self):
        ver = semver.VersionInfo.parse(self.specification.openapi)
        assert ver.major == 3
        assert ver.minor <= 0

        self.specification.info
        self.specification.paths

        schemas = self.specification.get("components", {}).get("schemas", [])
        for name, schema in schemas.items():
            self.add_schema(name, schema)

        with self as imp:
            for path in self.specification.paths:
                for name, fn in self.gen_functions(path).items():
                    imp.expose(name, fn)

    def add_schema(self, name: str, schema: Box):
        with self as imp:
            imp(name, self.typedef_from_jsonschema(schema))

    def gen_functions(self, path: str):
        return Path(self, path).generate_functions()


METHODS = (
    "get",
    "put",
    "post",
    "delete",
    "options",
    "head",
    "patch",
    "trace",
)


def get_success_code(responses: Box) -> Optional[str]:
    if "200" in responses:
        return "200"
    for status in responses:
        if re.match(r"^2\d\d$", status):
            return status
    # TODO what if there are more than one 2xx codes

    if "default" in responses:
        return "default"

    return None


class Path:
    path: str
    spec: Box
    runtime: str
    typedef_from_jsonschema: TypedefFromJsonSchema
    resolve_ref: Callable[[str], str]

    def __init__(self, importer: OpenApiImporter, path: str):
        self.path = path
        self.spec = importer.resolve_ref(importer.specification.paths[path])
        self.runtime = importer.name
        self.typedef_from_jsonschema = importer.typedef_from_jsonschema
        self.resolve_ref = lambda ref: importer.resolve_ref(ref)

    def generate_functions(self):
        ret = {}
        for method in self.spec:
            if method in METHODS:
                try:
                    name, fn = self.generate_function(method)
                    ret[name] = fn
                except Exception as e:
                    name = self.spec[method].operationId
                    print(
                        f"Warning: Generation of function '{name}' skipped: {e}",
                        file=stderr,
                    )
            else:
                print(
                    f"Warning: Unsupported method {method}: generation skipped",
                    file=stderr,
                )
        return ret

    def generate_function(self, method: str) -> Tuple[str, t.func]:
        spec = self.spec[method]
        inp_type, inp_config = self.gen_input_type(spec)
        success_code = get_success_code(spec.responses)

        if success_code is None:
            out = t.struct()
        elif success_code == "204":  # No Content
            out = t.boolean()
        else:
            res = spec.responses[success_code]
            if "content" in res:
                content = res.content
                if MIME_TYPES.json in content:
                    out = self.typedef_from_jsonschema(content[MIME_TYPES.json].schema)
                else:
                    raise Exception(
                        f"No supported content type for response: {', '.join(content)}"
                    )
            else:
                out = t.struct()

        # TODO what for other 4xx codes
        if "404" in spec.responses:
            out = out.optional()

        return (
            spec.operationId,
            t.func(
                inp_type,
                out,
                TypifyMat(
                    lambda inp, out: f"{self.runtime}.{method}('{self.path}', {inp}, {out}, {as_kwargs(inp_config)})"
                ),
            ),
        )

    def gen_input_type(self, op_spec: Box) -> Tuple[t.typedef, Box]:
        """Returns a tuple of t.typedef and a box of kwargs"""

        params = [*self.spec.get("parameters", []), *op_spec.get("parameters", [])]

        kwargs = Box({})
        props = {}
        for param in params:
            param = self.resolve_ref(param)
            if "schema" not in param:
                raise Exception(f"Unsupported parameter definition '{param}'")
            props[param.name] = self.typedef_from_jsonschema(param.schema)

        if "requestBody" in op_spec:
            body = op_spec.requestBody.content
            types = body.keys()
            if MIME_TYPES.json in types:
                content_type = MIME_TYPES.json
            elif MIME_TYPES.urlenc in types:
                content_type = MIME_TYPES.json
            elif MIME_TYPES.multipart in types:
                content_type = MIME_TYPES.multipart
            else:
                raise Exception(
                    f"No supported content type for request: {', '.join(types)}"
                )

            schema = body[content_type].schema
            schema = self.resolve_ref(schema)
            if schema.type != "object":
                raise Exception(f"Unsupported type for request body: {schema.type}")

            for name, typ in self.typedef_from_jsonschema(schema).props.items():
                if name in props:
                    raise Exception(
                        f"Name clash: '{name}' present in both query parameters and request body"
                    )
                props[name] = typ

            kwargs.content_type = repr(content_type)
            kwargs.body_fields = repr(tuple(schema.properties.keys()))

        return (t.struct(props), kwargs)


def as_kwargs(kwargs: Dict[str, str]):
    cg = ""
    for key, val in kwargs.items():
        cg += f"{key}={val},"
    return cg



# Copyright Metatype OÜ under the Elastic License 2.0 (ELv2). See LICENSE.md for usage.

from typing import TYPE_CHECKING, Callable, Optional, Union, cast

from attrs import define, field, frozen

from typegraph import effects
from typegraph import types as t
from typegraph.effects import Effect
from typegraph.graph.nodes import NodeProxy
from typegraph.runtimes.base import Materializer, Runtime
from typegraph.utils.attrs import always

if TYPE_CHECKING:
    from typegraph.importers.base.importer import Importer


@frozen
class TypifyRuntime(Runtime):
    runtime_name: str = always("__typify__")

    def data(self, _c):
        raise Exception("Cannot serialize TypifyRuntime")


@frozen
class TypifyMat(Materializer):
    codegen: Union[str, Callable[[str, str], str]]
    runtime: Runtime = field(factory=TypifyRuntime)
    effect: Effect = always(effects.none())

    def data(self, _c):
        raise Exception("Cannot serialize TypifyMat")


@define
class Typify:
    importer: "Importer"
    ns: str = field(default="t")

    def __call__(self, typ: t.TypeNode, name: Optional[str] = None) -> str:
        # dispatch
        if isinstance(typ, NodeProxy):
            renames = self.importer.renames
            name = renames[typ.node] if typ.node in renames else typ.node
            return f"{self.ns}.proxy(renames[{repr(name)}])"

        if hasattr(self, typ.type):
            method = getattr(self, typ.type)
        else:
            if typ.type in simple_types:
                method = getattr(self, "simple")
            else:
                raise Exception(f"No handler for type '{typ.type}'")

        suffix = "" if name is None else f".named(renames[{repr(name)}])"
        return method(typ) + suffix

    def constraints(typ: t.typedef) -> str:
        ret = ""
        if not hasattr(typ, "_constraints"):
            return ret
        for k, v in typ._constraints().items():
            ret += f".{k}({repr(v)})"
        return ret

    def simple(self, typ: t.typedef) -> str:
        return f"{self.ns}.{typ.type}(){Typify.constraints(typ)}"

    def optional(self, typ: t.typedef) -> str:
        typ = cast(t.optional, typ)
        ret = f"{self(typ.of)}.optional()"
        if typ.default_value is not None:
            ret += f".default({repr(typ.default_value)})"
        return ret

    def object(self, typ: t.typedef) -> str:
        typ = cast(t.struct, typ)
        fields = ", ".join([f"{repr(k)}: {self(v)}" for k, v in typ.props.items()])
        ret = f"{self.ns}.struct({{{fields}}})"
        if typ.additional_props is not None:
            if isinstance(typ.additional_props, bool):
                ret += f".additional({repr(typ.additional_props)})"
            else:
                ret += f".additional({self(typ.additional)})"
        ret += Typify.constraints(typ)
        return ret

    def array(self, typ: t.typedef) -> str:
        typ = cast(t.array, typ)
        return f"{self.ns}.array({self(typ.of)}){Typify.constraints(typ)}"

    def union(self, typ: t.typedef) -> str:
        typ = cast(t.union, typ)
        variants = [self(v) for v in typ.variants]
        return f"{self.ns}.union({', '.join(variants)})"

    def function(self, typ: t.typedef) -> str:
        typ = cast(t.func, typ)
        assert isinstance(typ.mat, TypifyMat)
        codegen = typ.mat.codegen
        if callable(codegen):
            return codegen(self(typ.inp), self(typ.out))
        return f"{self.ns}.func({self(typ.inp)}, {self(typ.out)}, {typ.mat.codegen}"


simple_types = {"boolean", "number", "integer", "string"}


def typify(typ: t.typedef, ns: str = "t") -> str:
    tp = Typify(ns)
    return tp(typ)



# Copyright Metatype OÜ under the Elastic License 2.0 (ELv2). See LICENSE.md for usage.

import inspect
from pathlib import Path
from typing import TYPE_CHECKING, Callable, Dict, List, Optional, Set, Union

from typegraph.graph.builder import Collector
from typegraph.graph.models import Auth, Cors, Rate
from typegraph.graph.nodes import Node, NodeProxy
from typegraph.runtimes.deno import DenoRuntime

if TYPE_CHECKING:
    from typegraph import types as t


OperationTable = Dict[str, Union["t.func", "t.struct"]]

typegraph_version = "0.0.1"


class TypeGraph:
    Auth = Auth
    Cors = Cors
    Rate = Rate

    # repesent domain projected to an interface
    name: str
    # version should be commit based
    # types: List["t.typedef"]
    type_by_names: Dict[str, "t.typedef"]  # for explicit names
    exposed: OperationTable
    latest_type_id: int
    auths: List[Auth]
    rate: Optional[Rate]
    cors: Cors
    path: str

    def __init__(
        self,
        name: str,
        auths=None,
        rate=None,
        cors=None,
    ) -> None:
        super().__init__()
        self.name = name
        # self.types = []
        self.type_by_names = {}
        self.exposed = {}
        self.policies = []
        self.latest_type_id = 0
        self.auths = [] if auths is None else auths
        self.rate = rate
        self.cors = Cors() if cors is None else cors
        self.path = Path(inspect.stack()[1].filename)

    def next_type_id(self):
        self.latest_type_id += 1
        return self.latest_type_id

    def __enter__(self):
        TypegraphContext.push(self)
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        TypegraphContext.pop()

    def __call__(
        self, node: str, after_apply: Callable[["t.typedef"], "t.typdef"] = lambda x: x
    ) -> "NodeProxy":
        return NodeProxy(self, node, after_apply)

    def expose(self, **ops: Union["t.func", "t.struct"]):
        from typegraph import types as t

        default_policy = ops.pop("default_policy", [])

        # allow to expose only functions or structures (namespaces)
        for name, op in ops.items():
            if not isinstance(op, t.func) and not isinstance(op, t.struct):
                raise Exception(
                    f"cannot expose type {op.title} under {name}, requires a function or structure (namespace), got a {op.type}"
                )

            if name in self.exposed:
                raise Exception(f"operation {name} already exposed")

            self.exposed[name] = op.add_policy(*default_policy)

        return self

    def root(self) -> "t.struct":
        from typegraph import types as t

        def assert_no_effect_materializers(
            tpe: Union[t.typedef, NodeProxy], history: Set[t.typedef] = set()
        ):
            if isinstance(tpe, NodeProxy):
                tpe = tpe.get()

            if tpe in history:
                return

            for e in tpe.edges:
                if not isinstance(e, t.typedef):
                    continue

                if isinstance(e, t.func):
                    if e.mat.effect.effect is not None:
                        raise Exception(
                            f"expected materializer to have no effect: ({e.mat})"
                        )
                    assert_no_effect_materializers(e.out)

                else:
                    assert_no_effect_materializers(e)

        with self:
            root = t.struct(self.exposed).named(self.name)

        root._propagate_runtime(DenoRuntime())

        return root

    def collect_nodes(self) -> Collector:
        def visit(nodes: List[Node], collector: Collector):
            for node in nodes:
                if collector.collect(node):
                    visit(node.edges, collector)

        collector = Collector()
        visit([self.root()], collector)

        return collector

    def build(self):
        if self.name.startswith("__"):
            raise Exception(f"cannot build typegraph with reserved name '{self.name}'")

        collector = self.collect_nodes()

        ret = {
            c: [n.data(collector) for n in collector.collects[c]]
            for c in collector.collects
        }

        ret["meta"] = {
            "secrets": ret.pop("secrets") if "secrets" in ret else [],
            "auths": self.auths,
            "rate": self.rate,
            "cors": self.cors,
            "version": typegraph_version,
        }

        ret["$id"] = f"https://metatype.dev/specs/{typegraph_version}.json"

        return ret


class TypegraphContext:
    typegraphs: List[TypeGraph] = []

    @classmethod
    def push(cls, tg: TypeGraph):
        cls.typegraphs.append(tg)

    @classmethod
    def pop(cls) -> Optional[TypeGraph]:
        try:
            return cls.typegraphs.pop()
        except IndexError:
            return None

    @classmethod
    def get_active(cls) -> Optional[TypeGraph]:
        try:
            return cls.typegraphs[-1]
        except IndexError:
            return None


def get_absolute_path(relative: str) -> Path:
    tg_path = TypegraphContext.get_active().path
    return tg_path.parent / relative


def find(node: str) -> Optional["t.typedef"]:
    g = TypegraphContext.get_active()
    if g is None:
        raise Exception("No active TypegraphContext")
    if node in g.type_by_names:
        return g.type_by_names[node]
    else:
        return None


def resolve_proxy(tpe: Union[NodeProxy, "t.typedef"]) -> "t.typedef":
    from typegraph import types as t

    if isinstance(tpe, NodeProxy):
        return tpe.get()
    else:
        assert isinstance(tpe, t.typedef)
        return tpe



# Copyright Metatype OÜ under the Elastic License 2.0 (ELv2). See LICENSE.md for usage.

from typing import Callable, Dict, List, Optional

from attrs import field, frozen

from typegraph import effects
from typegraph import types as t
from typegraph import utils
from typegraph.effects import Effect
from typegraph.runtimes.base import Materializer, Runtime
from typegraph.utils.attrs import always


def pick(d: Dict, *largs) -> Dict:
    return utils.drop_nones(utils.pick(d, *largs))


@frozen
class RandomRuntime(Runtime):
    runtime_name: str = always("random")
    seed: Optional[int] = field(kw_only=True, default=None)
    reset: str = field(kw_only=True, default="")

    def get_type_config(self, tpe: "t.typedef") -> Dict:
        base = tpe.runtime_config

        def pick(*largs) -> Dict:
            return utils.drop_nones(utils.pick(base, *largs))

        gen = base.get("gen")

        def pick_fields(default: Callable[[], Dict] = dict, **d: List[str]) -> Dict:
            if gen in d:
                return pick("gen", *d[gen])
            else:
                return default()

        if tpe.type == "boolean":
            return pick(base, "likelihood")

        if tpe.type == "number":
            if gen == "float":
                return pick("gen", "fixed", "min", "max")

        if tpe.type == "integer":
            if gen == "age":
                return pick("gen", "type")
            if gen == "natural" or gen == "prime" or gen is None:
                return pick("gen", "min", "max")

        if tpe.type == "string":
            return pick_fields(
                (
                    lambda: pick(
                        "length", "pool", "alpha", "numeric", "casing", "symbols"
                    )
                ),
                char=["pool", "alpha", "numeric", "casing", "symbols"],
                letter=["casing"],
                paragraph=["sentences"],
                sentence=["words"],
                syllable=[],
                word=["length", "syllables"],
                birthday=[],
                first=["nationality", "gender"],
                last=["nationality", "gender"],
                name=["middle", "middle_initial", "prefix", "nationality"],
                gender=["extraGenders"],
                prefix=["full", "gender"],
                suffix=["full"],
                animal=["type"],
                address=["short_suffix"],
                city=[],
                country=["full"],
                postcode=[],
            )
            # TODO more...

        return dict()

    def generate(self, out, **kwargs):
        return t.gen(out, RandomMat(runtime=self), **kwargs)


@frozen
class RandomMat(Materializer):
    runtime: Runtime = field(factory=RandomRuntime)
    materializer_name: str = always("random")
    effect: Effect = field(kw_only=True, default=effects.none())



# Copyright Metatype OÜ under the Elastic License 2.0 (ELv2). See LICENSE.md for usage.

from typegraph import types as t
from typegraph.graph.nodes import NodeProxy


def resolve_entity_quantifier(tpe: t.typedef):
    if isinstance(tpe, t.array):
        return tpe.of
    if isinstance(tpe, t.optional):
        return tpe.of
    return tpe


def clean_virtual_link(tpe: t.typedef):
    if isinstance(tpe, t.struct):
        ret = {}
        # renames = {}
        for k, v in tpe.props.items():
            if isinstance(v, NodeProxy):
                v = v.get()

            if isinstance(v, t.func):
                if isinstance(v.out, t.array) and isinstance(v.out.of, t.struct):
                    continue

                # ids = v.inp.of.keys()
                # key = "_".join(ids)
                # if len(ids) <= 1:
                #    key = f"{k}_{key}"
                # renames[k] = key

                out = clean_virtual_link(v.out)
                ret[k] = t.struct(
                    {
                        "create": out.optional(),
                        "createMany": t.struct({"data": t.array(out)}).optional(),
                        "connect": out.optional(),
                        "connectOrCreate": t.struct(
                            {
                                # "where":
                                "create": out,
                            }
                        ).optional(),
                    }
                ).optional()
            else:
                ret[k] = clean_virtual_link(v)

        ret = t.struct(ret)
        # ret.renames = renames
        return ret

    return tpe


def only_unique(tpe: t.Type):
    if isinstance(tpe, t.struct):
        return t.struct({k: only_unique(v) for k, v in tpe.ids().items()})
    return tpe


def optional_root(tpe: t.struct):
    return t.struct({k: v.optional() for k, v in tpe.of.items()})



# Copyright Metatype OÜ under the Elastic License 2.0 (ELv2). See LICENSE.md for usage.

from typing import Optional, Tuple

from attrs import field, frozen

from typegraph import effects
from typegraph import types as t
from typegraph.effects import Effect
from typegraph.runtimes.base import Materializer, Runtime
from typegraph.utils.attrs import always, required


@frozen
class HTTPRuntime(Runtime):
    """Runs HTTP requests.

    Example:
    ```python
    from typegraph.runtime.http import HTTPRuntime

    remote = HTTPRuntime('https://dev.to/api')
    remote.get(
        '/test',
        t.struct({}),
        t.array(t.struct({'a': t.integer()})),
    )
    ```
    """

    endpoint: str
    cert_secret: Optional[str] = None
    basic_auth_secret: Optional[str] = None
    runtime_name: str = always("http")

    def data(self, collector):
        return {
            **super().data(collector),
            "cert_secret": self.cert_secret,
            "basic_auth_secret": self.basic_auth_secret,
        }

    def get(self, path: str, inp, out, effect: Effect = effects.none(), **kwargs):
        return t.func(
            inp,
            out,
            RESTMat(self, "GET", path, effect=effect, **kwargs),
        )

    def post(self, path: str, inp, out, effect: Effect = effects.create(), **kwargs):
        return t.func(
            inp,
            out,
            RESTMat(self, "POST", path, effect=effect, **kwargs),
        )

    def put(self, path: str, inp, out, effect: Effect = effects.upsert(), **kwargs):
        return t.func(
            inp,
            out,
            RESTMat(self, "PUT", path, effect=effect, **kwargs),
        )

    def patch(self, path: str, inp, out, effect: Effect = effects.update(), **kwargs):
        return t.func(
            inp,
            out,
            RESTMat(self, "PATCH", path, effect=effect, **kwargs),
        )

    def delete(self, path: str, inp, out, effect: Effect = effects.delete(), **kwargs):
        return t.func(
            inp,
            out,
            RESTMat(self, "DELETE", path, effect=effect, **kwargs),
        )


@frozen
class RESTMat(Materializer):
    runtime: Runtime
    verb: str
    path: str

    content_type: str = field(kw_only=True, default="application/json")
    header_prefix: Optional[str] = "header#"
    query_fields: Optional[Tuple[str, ...]] = field(kw_only=True, default=None)
    body_fields: Optional[Tuple[str, ...]] = field(kw_only=True, default=None)
    auth_token_field: Optional[str] = field(kw_only=True, default=None)
    materializer_name: str = always("rest")
    effect: Effect = required()



# Copyright Metatype OÜ under the Elastic License 2.0 (ELv2). See LICENSE.md for usage.

from attrs import frozen

from typegraph import effects, t
from typegraph.effects import Effect
from typegraph.runtimes.base import Materializer, Runtime
from typegraph.utils.attrs import always


@frozen
class S3Runtime(Runtime):
    host: str
    region: str
    access_key_secret: str
    secret_key_secret: str
    runtime_name: str = always("s3")

    def data(self, collector):
        data = super().data(collector)
        data["data"].update(
            host=self.host,
            region=self.region,
            access_key_secret=self.access_key_secret,
            secret_key_secret=self.secret_key_secret,
        )
        return data

    def sign(self, bucket: str, content_type: str):
        return t.func(
            t.struct({"length": t.integer(), "path": t.string()}),
            t.string(),
            SignMat(self, bucket, content_type),
        )

    def list(self, bucket: str):
        return t.func(
            t.struct({"path": t.string()}),
            t.struct(
                {
                    "keys": t.array(t.struct({"key": t.string(), "size": t.integer()})),
                    "prefix": t.array(t.string()),
                }
            ),
            ListMat(self, bucket),
        )


@frozen
class SignMat(Materializer):
    runtime: Runtime
    bucket: str
    content_type: str
    materializer_name: str = always("sign")
    effect: Effect = always(effects.none())


@frozen
class ListMat(Materializer):
    runtime: Runtime
    bucket: str
    materializer_name: str = always("list")
    effect: Effect = always(effects.none())



# Copyright Metatype OÜ under the Elastic License 2.0 (ELv2). See LICENSE.md for usage.

from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    Union,
    get_args,
    get_origin,
)

from attrs import evolve, field, frozen
from frozendict import frozendict
from typing_extensions import Self

from typegraph.graph.builder import Collector
from typegraph.graph.nodes import Node, NodeProxy
from typegraph.graph.typegraph import TypeGraph, TypegraphContext, find
from typegraph.policies import Policy
from typegraph.runtimes.base import Materializer, Runtime
from typegraph.utils.attrs import SKIP, always, asdict

# if os.environ.get("DEBUG"):
#     import debugpy

#     debugpy.listen(5678)
#     print("Waiting for debugger attach...")
#     debugpy.wait_for_client()


TypeNode = Union["typedef", NodeProxy]


def remove_none_values(obj):
    return {k: v for k, v in obj.items() if v is not None}


def is_optional(tpe: Type):
    # Optional = Union[T, NoneType]
    return get_origin(tpe) is Union and type(None) in get_args(tpe)


class Secret(Node):
    secret: str

    def __init__(self, secret: str):
        super().__init__("secrets")
        self.secret = secret

    def data(self, collector: "Collector") -> dict:
        return self.secret  # this is a string


# Optional keyword-only field
def optional_field():
    return field(kw_only=True, default=None)


# reserved types are used for internal implementation
# and cannot be used to define user custom types
#
# more information at https://spec.graphql.org/draft/
reserved_types = [
    # Primitives
    "Int",
    "Float",
    "Boolean",
    "String",
    # Operations
    "Query",
    "Mutation",
    "Subscription",
    "Fragment",
    # Extras
    "Date",
    "ID",
    "UUID",
    "URL",
    "Point",
    "PointList",
    "Polygon",
    "MultiPolygon",
]


@frozen
class typedef(Node):
    graph: TypeGraph = field(
        kw_only=True,
        factory=TypegraphContext.get_active,
        init=False,
        metadata={SKIP: True},
    )
    name: str = field(kw_only=True, default="")
    description: Optional[str] = optional_field()
    runtime: Optional["Runtime"] = optional_field()
    inject: Optional[Union[str, TypeNode]] = optional_field()
    injection: Optional[Any] = optional_field()
    policies: Tuple[Policy, ...] = field(kw_only=True, factory=tuple)
    # runtime_config: Dict[str, Any] = field(
    #     kw_only=True, factory=dict, hash=False, metadata={SKIP: True}
    # )
    runtime_config: Dict[str, Any] = field(
        kw_only=True, factory=frozendict, hash=False, metadata={SKIP: True}
    )
    _enum: Optional[Tuple[Any]] = optional_field()

    collector_target: Optional[str] = always(Collector.types)

    def __attrs_post_init__(self):
        if self.graph is None:
            raise Exception("No typegraph context")
        if self.name == "":
            object.__setattr__(self, "name", f"{self.type}_{self.graph.next_type_id()}")
        if self.inject is None:
            object.__setattr__(self, "injection", None)

    def replace(self, **changes):
        return evolve(self, **changes)

    @property
    def type(self):
        return type(self).__name__

    @property
    def edges(self) -> List[Node]:
        secret = self.inject if self.injection == "secret" else None
        return (
            super().edges
            + list(self.policies)
            + list(filter(None, [self.runtime, secret]))
        )

    @property
    def labeled_edges(self) -> Dict[str, str]:
        return {}

    def register_name(self):
        types = self.graph.type_by_names
        name = self.name
        if name in types:
            if types[name] != self:
                raise Exception(f"Name '{name}' is already registered for another type")
            return

        if name in reserved_types:
            raise Exception(f"Type name '{name}' is a reserved type")
        # https://spec.graphql.org/draft/#sel-GAJTBAABABFj6D
        if name.startswith("__"):
            raise Exception(
                f"type name {name} cannot start with `__`, it's reserved for introspection"
            )
        types[name] = self

    def named(self, name: str) -> "typedef":
        ret = self.replace(name=name)
        ret.register_name()
        return ret

    def describe(self, description: str) -> "typedef":
        return self.replace(description=description)

    def within(self, runtime):
        if runtime is None:
            raise Exception("cannot set runtime to None")

        if self.runtime is not None and self.runtime != runtime:
            raise Exception(
                f"trying to set a different runtime {runtime}, already is {self.runtime}"
            )

        object.__setattr__(self, "runtime", runtime)
        return self

    def _propagate_runtime(self, runtime: "Runtime", visited: Set["typedef"] = None):
        if visited is None:
            visited = set()
        elif self in visited:
            return
        else:
            visited.add(self)

        if self.runtime is None:
            object.__setattr__(self, "runtime", runtime)

        for e in self.edges:
            if isinstance(e, typedef):
                e._propagate_runtime(self.runtime, visited)
            if isinstance(e, NodeProxy):
                e.get()._propagate_runtime(self.runtime, visited)

    def set(self, value):
        if self.inject is not None:
            raise Exception(f"{self.name} can only have one injection")

        import json

        return self.replace(injection="raw", inject=json.dumps(value))

    def from_secret(self, secret_name):
        if self.inject is not None:
            raise Exception(f"{self.name} can only have one injection")

        return self.replace(injection="secret", inject=Secret(secret_name))

    def from_parent(self, sibiling: "NodeProxy"):
        # TODO: check for same type and value in same context
        if self.inject is not None:
            raise Exception(f"{self.name} can only have one injection")

        return self.replace(injection="parent", inject=sibiling)

    def from_context(self, claim: str):
        if self.inject is not None:
            raise Exception(f"{self.name} can only have one injection")

        return self.replace(injection="context", inject=claim)

    def add_policy(
        self,
        *policies: List[Union[Policy, Materializer]],
    ):
        return self.replace(
            policies=self.policies + tuple(Policy.get_from(p) for p in policies)
        )

    def config(self, *flags: str, **kwargs: Any):
        d = dict()
        d.update(self.runtime_config)
        d.update(kwargs)
        d.update({f: True for f in flags})
        return self.replace(runtime_config=frozendict(d))

    def enum(self, variants: List[Any]) -> Self:
        return self.replace(enum=tuple(variants))

    def data(self, collector) -> dict:
        if self.runtime is None:
            raise Exception("Expected Runtime, got None")

        ret = remove_none_values(asdict(self))
        ret["type"] = self.type
        ret["title"] = ret.pop("name")
        ret["runtime"] = collector.index(self.runtime)
        ret["policies"] = [collector.index(p) for p in self.policies]
        if self.injection == "parent":
            ret["inject"] = collector.index(self.inject)
        elif self.injection == "secret":
            ret["inject"] = self.inject.secret

        config = self.runtime.get_type_config(self)
        if len(config) > 0:
            ret["config"] = config

        if hasattr(type(self), "constraint_data"):
            ret.update(self.constraint_data())

        if self._enum is not None:
            ret["enum"] = ret.pop("_enum")

        ret.pop("collector_target")

        return ret

    def optional(self, default_value: Optional[Any] = None) -> "optional":
        if isinstance(self, optional):
            return self

        with self.graph:
            return optional(self, default_value=default_value)

    def __repr__(self) -> str:
        def pretty(v):
            if isinstance(v, frozendict):
                return dict(v)
            return v

        attrs = ", ".join(
            f"{k} → {pretty(v)}"
            for k, v in vars(self).items()
            if k not in ("graph",) and v is not None
        )
        return f"{self.type}({attrs})"


#
# Type constraints

TYPE_CONSTRAINT = "__type_constraint_name"


def constraint(name: Optional[str] = None):
    # Additional constraint on type: Validation keyword.
    # Field to be manually set on the serialization.
    return field(
        kw_only=True, default=None, metadata={SKIP: True, TYPE_CONSTRAINT: name or True}
    )


def with_constraints(cls):
    if not hasattr(cls, "__attrs_attrs__"):
        raise Exception(
            "@with_constraints decorator requires class to have attribute '__attrs_attrs__'"
        )
    if not issubclass(cls, typedef):
        raise Exception("@with_constraints decorator requires a subclass of 'typedef'")

    constraints = {}

    def get_setter(name: str):
        def setter(self, value):
            return self.replace(**{name: value})

        return setter

    for f in cls.__attrs_attrs__:
        if f.metadata is not None and TYPE_CONSTRAINT in f.metadata:
            if not f.name.startswith("_"):
                raise Exception(
                    f"constraint field name '{f.name}' expected to start with an underscore"
                )
            name = f.name[1:]
            key = f.metadata.get(TYPE_CONSTRAINT)
            if isinstance(key, bool) and key:
                key = name
            constraints[key] = f.name

            setattr(cls, name, get_setter(name))

    if len(constraints) > 0:

        def constraint_data(self):
            return remove_none_values(
                {key: getattr(self, name) for key, name in constraints.items()}
            )

        def _constraints(self):
            return remove_none_values(
                {name[1:]: getattr(self, name) for name in constraints.values()}
            )

        setattr(cls, "constraint_data", constraint_data)
        setattr(cls, "_constraints", _constraints)

    return cls


# end - Type constraints


@frozen
class optional(typedef):
    of: TypeNode
    default_value: Optional[str] = field(hash=False, default=None)

    def default(self, value):
        if self.default_value is not None:
            raise Exception(f"{self.name} can only have one injection")

        return self.replace(default_value=value)

    @property
    def edges(self) -> List[Node]:
        return super().edges + [self.of]

    @property
    def labeled_edges(self) -> Dict[str, str]:
        return {"[item]": self.of.name}

    def data(self, collector) -> dict:
        ret = super().data(collector)
        ret["item"] = collector.index(ret.pop("of"))
        return ret


@frozen
class boolean(typedef):
    pass


@with_constraints
@frozen
class number(typedef):
    """Represents a generic number.

    Args:
        _min (float, optional): minimum constraint
        _max (float, optional): maximum constraint
        _x_min (float, optional): exclusive minimum constraint
        _x_max (float, optional): exclusive maximum constraint
        _multiple_of (float, optional): number must be a multiple of
    """

    _min: Optional[float] = constraint("minimum")
    _max: Optional[float] = constraint("maximum")
    _x_min: Optional[float] = constraint("exclusiveMinimum")
    _x_max: Optional[float] = constraint("exclusiveMaximum")
    _multiple_of: Optional[float] = constraint("multipleOf")


def float() -> number:
    return number()


@frozen
class integer(number):
    """An integer."""

    pass


@with_constraints
@frozen
class string(typedef):
    _min: Optional[int] = constraint("minLength")
    _max: Optional[int] = constraint("maxLength")
    _pattern: Optional[str] = constraint()
    _format: Optional[str] = constraint()

    def uuid(self) -> "string":
        return self.format("uuid")

    def email(self) -> "string":
        return self.format("email")

    def uri(self) -> "string":
        return self.format("uri")

    def json(self) -> "string":
        return self.format("json")

    def hostname(self) -> "string":
        return self.format("hostname")


def uuid() -> string:
    return string().uuid()


def email() -> string:
    return string().email()


def uri() -> string:
    return string().uri()


def json() -> string:
    return string().json()


def ean() -> string:
    return string().format("ean")


def path() -> string:
    return string().format("path")


def datetime() -> string:
    return string().format("date-time")


def date() -> string:
    return string().format("date")


def phone() -> string:
    # TODO replace with the phone number pattern when typechecking
    return string().format("phone")


def enum(variants: List[str]) -> string:
    return string().enum(variants)


def validate_struct_props(instance, attribute, props):
    for tpe in props.values():
        if not isinstance(tpe, typedef) and not isinstance(tpe, NodeProxy):
            raise Exception(
                f"expected typedef or NodeProxy, got {type(tpe).__name__}: {props}"
            )


@with_constraints
@frozen
class struct(typedef):
    props: Dict[str, TypeNode] = field(
        validator=[validate_struct_props], factory=frozendict, converter=frozendict
    )
    additional_props: Optional[Union[bool, TypeNode]] = None
    _min: Optional[int] = constraint("minProperties")
    _max: Optional[int] = constraint("maxProperties")
    # _dependentRequired

    def additional(self, t: Union[bool, TypeNode]):
        return self.replace(additional_props=t)

    def compose(self, props: Dict[str, typedef]):
        new_props = dict(self.props)
        new_props.update(props)
        return self.replace(props=frozendict(new_props))

    def __getattr__(self, attr):
        try:
            return super().__getattr__(attr)
        except AttributeError:
            pass

        if attr in self.props:
            return self.props[attr]

        raise Exception(f'no prop named "{attr}" in type {self}')

    @property
    def type(self):
        return "object"

    @property
    def edges(self) -> List[Node]:
        return super().edges + list(self.props.values())

    @property
    def labeled_edges(self) -> Dict[str, str]:
        return {k: v.name for k, v in self.props.items()}

    def data(self, collector) -> dict:
        ret = super().data(collector)
        if self.additional_props is not None:
            ret["additional_props"] = (
                collector.index(self.additional_props)
                if isinstance(self.additional_props, Node)
                else self.additional_props
            )
        ret["properties"] = {k: collector.index(v) for k, v in ret.pop("props").items()}
        return ret


def any_object() -> struct:
    return struct().additional(True)


@with_constraints
@frozen
class array(typedef):
    of: TypeNode

    _min: Optional[int] = constraint("minItems")
    _max: Optional[int] = constraint("maxItems")
    _unique_items: Optional[bool] = constraint("uniqueItems")
    # _min_contains: Optional[int] = None
    # _max_contains: Optional[int] = None

    @property
    def edges(self) -> List[Node]:
        return super().edges + [self.of]

    @property
    def labeled_edges(self):
        return {"[items]": self.of.name}

    def data(self, collector) -> dict:
        ret = super().data(collector)
        ret["items"] = collector.index(ret.pop("of"))
        return ret


@frozen
class union(typedef):
    """A `union` type represents a general union with the variants provided.

    The `union` type is equivalent to the `anyOf` field in JSON Schema where
    the given data must be valid against one or more of the given subschemas.
    """

    variants: List[TypeNode] = list()

    def __hash__(self):
        return hash(tuple(self.variants))

    @property
    def edges(self) -> List[Node]:
        nodes = super().edges + list(self.variants)
        return nodes

    def data(self, collector: Collector) -> dict:
        ret = super().data(collector)
        ret["anyOf"] = [collector.index(v) for v in ret.pop("variants")]
        return ret


@frozen
class either(typedef):
    """An `either` type represents a disjoint union with the variants provided.

    The `either` type is equivalent to the `oneOf` field in JSON Schema where
    the given data must be valid against exactly one of the given subschemas.
    """

    variants: List[TypeNode] = list()

    def __hash__(self):
        return hash(tuple(self.variants))

    @property
    def edges(self) -> List[Node]:
        nodes = super().edges + list(self.variants)
        return nodes

    def data(self, collector: Collector) -> dict:
        ret = super().data(collector)
        ret["oneOf"] = [collector.index(v) for v in ret.pop("variants")]
        return ret


def ipv4() -> string:
    return string().format("ipv4")


def ipv6() -> string:
    return string().format("ipv6")


def ip() -> union:
    return union([ipv4(), ipv6()])


@frozen
class any(typedef):
    pass


@frozen
class func(typedef):
    inp: TypeNode
    out: TypeNode
    mat: Materializer

    # if not safe, output will be typechecked
    safe: bool = field(kw_only=True, default=True)

    rate_calls: bool = field(kw_only=True, default=False)
    rate_weight: Optional[int] = field(kw_only=True, default=None)

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        object.__setattr__(self, "runtime", self.mat.runtime)

    def rate(self, weight=None, calls=False):
        return self.replace(rate_weight=weight, rate_calls=calls)

    @property
    def edges(self) -> List[Node]:
        return super().edges + [self.inp, self.out, self.mat]

    @property
    def labeled_edges(self) -> Dict[str, str]:
        return {"[in]": self.inp.name, "[out]": self.out.name}

    @property
    def type(self) -> str:
        return "function"

    def data(self, collector) -> dict:
        ret = super().data(collector)
        inp = ret.pop("inp")
        if isinstance(inp, NodeProxy):
            inp = inp.get()
        if not isinstance(inp, struct):
            raise Exception(f"invalid input type, expected struct got {inp.name}")
        ret["input"] = collector.index(inp)
        ret["output"] = collector.index(ret.pop("out"))
        ret["materializer"] = collector.index(ret.pop("mat"))
        return ret

    # this is not function composition
    def compose(self, out: Dict[str, typedef]):
        if not isinstance(self.out, struct):
            raise Exception("Output required to be a struct.")
        return self.replace(out=self.out.compose(out))

    # def compose(self, other: "func") -> "func":
    #     assert self.out == other.inp
    #     # what if other == gen?
    #     return func(self, other, IdentityMat())

    # def __mul__(self, other: "func") -> "func":
    #     return self.compose(other)


def gen(out: typedef, mat: Materializer, **kwargs) -> func:
    return func(struct(), out, mat, **kwargs)


# single instance
def named(name: str, define: Callable[[], typedef]) -> TypeNode:
    defined = find(name)
    if defined is not None:
        return defined
    else:
        return define().named(name)


def proxy(name: str) -> NodeProxy:
    return NodeProxy(TypegraphContext.get_active(), name)


def visit_reversed(
    tpe: typedef, fn: Callable[[typedef], Any], visited: Set[str] = set()
):
    for node in tpe.edges:
        if isinstance(node, typedef):
            if node.name in visited:
                continue
            visited.add(node.name)
            visit_reversed(node, fn, visited)
            fn(node)



# Copyright Metatype OÜ under the Elastic License 2.0 (ELv2). See LICENSE.md for usage.

from typing import Callable, Dict, List

from attrs import field, frozen
from box import Box
from deepmerge import always_merger

from typegraph import t

# TODO support for boolean schemas


@frozen
class TypedefFromJsonSchema:
    ref_to_name: Callable[[str], str]
    resolve_ref: Callable[[str], Box]
    type_dispatch: Dict[str, Callable[[Box], t.typedef]] = field(
        init=False, factory=dict
    )
    # resolve_ref: Callable[[str], str]

    def __attrs_post_init__(self):
        object.__setattr__(self, "type_dispatch", self._type_dispatch())

    def _type_dispatch(self):
        return {
            "integer": lambda _s: t.integer(),
            "number": lambda _s: t.float(),
            "string": lambda _s: t.string(),
            "boolean": lambda _s: t.boolean(),
            "array": lambda s: t.array(self(s["items"])),
            "object": lambda s: t.struct(
                {
                    prop_name: (
                        self(prop_schema)
                        if prop_name in s.get("required", [])
                        else self(prop_schema).optional()
                    )
                    for prop_name, prop_schema in s.get("properties", dict()).items()
                }
            ),
        }

    def __call__(self, schema: Box):
        if "$ref" in schema:
            return t.proxy(self.ref_to_name(schema["$ref"]))

        if "allOf" in schema:
            return self(Box(merge_all([self.resolve_ref(s) for s in schema.allOf])))

        if "type" not in schema:
            raise Exception(f"Unsupported schema: {schema}")

        gen = self.type_dispatch.get(schema.type)
        if gen is None:
            raise Exception(f"Unsupported type '{schema.type}'")

        return gen(schema)


def merge_all(items: List[dict]):
    ret = {}
    for item in items:
        always_merger.merge(ret, item)
    return ret



# Copyright Metatype OÜ under the Elastic License 2.0 (ELv2). See LICENSE.md for usage.

from attrs import field, frozen

from typegraph.effects import Effect
from typegraph.runtimes.base import Materializer, Runtime
from typegraph.utils.attrs import always, required


@frozen
class GoogleapisRuntime(Runtime):
    runtime_name: str = always("googleapis")


@frozen
class RestMat(Materializer):
    verb: str
    url: str
    effect: Effect = required()
    runtime: Runtime = field(kw_only=True, factory=GoogleapisRuntime)
    materializer_name: str = always("restmat")



# Copyright Metatype OÜ under the Elastic License 2.0 (ELv2). See LICENSE.md for usage.

from re import sub as reg_sub
from typing import List

from attrs import evolve, field, frozen

from typegraph.graph.builder import Collector
from typegraph.graph.nodes import Node
from typegraph.graph.typegraph import TypegraphContext
from typegraph.runtimes.base import Materializer
from typegraph.runtimes.deno import PureFunMat
from typegraph.utils.attrs import always
from typegraph.utils.sanitizers import sanitize_ts_string


def policy_name_factory():
    tg = TypegraphContext.get_active()
    if tg is None:
        raise Exception("typegraph context needed")
    return f"policy_{tg.next_type_id()}"


@frozen
class Policy(Node):
    name: str = field(factory=policy_name_factory, kw_only=True)
    mat: Materializer
    collector_target: str = always(Collector.policies)

    def named(self, name: str):
        return evolve(self, name=name)

    @property
    def edges(self) -> List[Node]:
        return [self.mat]

    def data(self, collector):
        return {
            "name": self.name,
            "materializer": collector.index(self.mat),
        }

    @classmethod
    def get_from(cls, p) -> "Policy":
        if isinstance(p, Materializer):
            return cls(p)
        if isinstance(p, cls):
            return p
        raise Exception(f"Cannot create Policy from a {type(p).__name__}")


def public(name: str = "__public"):
    return Policy(PureFunMat("() => true")).named(name)


def jwt(role_name: str, field: str = "role"):
    # join role_name, field with '__jwt' as prefix
    separator, pattern = "_", r"[^a-zA-Z0-9_]+"
    prefix = "__jwt"
    jwt_name = reg_sub(pattern, separator, separator.join([prefix, role_name, field]))
    # build the policy
    role_name = sanitize_ts_string(role_name)
    field = sanitize_ts_string(field)
    src = f"""
            (_, {{ context }}) => {{
                const role_chunks = "{role_name}".split(".");
                let value = context?.[role_chunks.shift()];
                for (const chunk of role_chunks)
                    value = value?.[chunk];
                return value === "{field}";
            }}
        """
    return Policy(PureFunMat(src)).named(jwt_name)


def internal():
    # https://metatype.atlassian.net/browse/MET-107
    return Policy(PureFunMat("() => false")).named("internal")



# Copyright Metatype OÜ under the Elastic License 2.0 (ELv2). See LICENSE.md for usage.

from typegraph.graph.models import Auth

github_auth = Auth.oauth2(
    "github",
    "https://github.com/login/oauth/authorize",
    "https://github.com/login/oauth/access_token",
    "openid profile email",
    "https://api.github.com/user",
)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['typegraph',
 'typegraph.graph',
 'typegraph.graph.auth',
 'typegraph.importers',
 'typegraph.importers.base',
 'typegraph.providers.aws.runtimes',
 'typegraph.providers.google.runtimes',
 'typegraph.providers.prisma',
 'typegraph.providers.prisma.runtimes',
 'typegraph.providers.temporal.runtimes',
 'typegraph.runtimes',
 'typegraph.utils']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'astunparse>=1.6.3,<2.0.0',
 'attrs>=22.2.0,<23.0.0',
 'black>=22.12,<24.0',
 'deepmerge>=1.1.0,<2.0.0',
 'frozendict==2.3.5',
 'graphql-core>=3.2.3,<4.0.0',
 'httpx[http2]>=0.22,<0.24',
 'ordered-set>=4.1.0,<5.0.0',
 'python-box>=7.0.0,<8.0.0',
 'redbaron>=0.9.2,<0.10.0',
 'semver>=2.13.0,<3.0.0',
 'strenum>=0.4.9,<0.5.0',
 'typing-extensions>=4.5.0,<5.0.0']

entry_points = \
{'console_scripts': ['py-tg = typegraph.utils.loaders:cmd']}

setup_kwargs = {
    'name': 'typegraph',
    'version': '0.0.3.dev5',
    'description': 'Free and open ecosystem for API composition.',
    'long_description': 'None',
    'author': 'Metatype Contributors',
    'author_email': 'support@metatype.dev',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://metatype.dev',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
